from __future__ import annotations

import itertools
import shutil
import sys
from types import ModuleType, SimpleNamespace
from typing import Any, cast

import click
from configobj import ConfigObj
import prompt_toolkit
from prompt_toolkit.formatted_text import ANSI, FormattedText, to_plain_text
import pytest

from mycli import compat
from mycli import output as output_module
from mycli.output import OutputMixin
from mycli.packages.sqlresult import SQLResult
from mycli.types import ImageProtocol
from test.utils import DummyFormatter, FakeCursorBase, make_bare_mycli  # type: ignore[attr-defined]


def test_output_timing_logs_and_prints_with_default_style(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    logged: list[Any] = []
    printed: list[tuple[Any, Any]] = []
    cli.log_output = lambda value: logged.append(value)  # type: ignore[assignment]
    monkeypatch.setattr(prompt_toolkit, 'print_formatted_text', lambda text, style=None: printed.append((text, style)))

    OutputMixin.output_timing(cli, '0.12 sec')

    assert logged == ['0.12 sec']
    assert to_plain_text(printed[0][0]) == '0.12 sec'
    assert list(printed[0][0])[0][0].strip() == 'class:output.timing'
    assert printed[0][1] == cli.ptoolkit_style


def test_output_timing_uses_warning_style(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.log_output = lambda value: None  # type: ignore[assignment]
    printed: list[Any] = []
    monkeypatch.setattr(prompt_toolkit, 'print_formatted_text', lambda text, style=None: printed.append(text))

    OutputMixin.output_timing(cli, '0.34 sec', is_warnings_style=True)

    assert list(printed[0])[0][0].strip() == 'class:warnings.timing'


def test_log_query_and_log_output_write_plain_text(tmp_path) -> None:
    cli = make_bare_mycli()
    logfile = tmp_path / 'audit.log'

    with logfile.open('w+', encoding='utf-8') as handle:
        cli.logfile = handle
        OutputMixin.log_query(cli, 'select 1')
        OutputMixin.log_output(cli, ANSI('\x1b[31mhello\x1b[0m'))
        handle.seek(0)
        contents = handle.read()

    assert 'select 1' in contents
    assert 'hello' in contents
    assert '\x1b[31m' not in contents


def test_log_output_ignores_missing_logfile() -> None:
    cli = make_bare_mycli()
    cli.logfile = None

    OutputMixin.log_output(cli, 'nothing to write')


def test_echo_logs_and_prints(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    logged: list[str] = []
    printed: list[tuple[str, dict[str, Any]]] = []
    cli.log_output = lambda value: logged.append(value)  # type: ignore[assignment]
    monkeypatch.setattr(click, 'secho', lambda value, **kwargs: printed.append((value, kwargs)))

    OutputMixin.echo(cli, 'message', fg='red')

    assert logged == ['message']
    assert printed == [('message', {'fg': 'red'})]


def test_output_emits_iterm2_image_without_text_sinks(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.explicit_pager = False
    cli.get_output_margin = lambda status=None: 1  # type: ignore[assignment]
    logged: list[Any] = []
    emitted: list[str] = []
    cli.log_output = lambda value: logged.append(value)  # type: ignore[assignment]
    monkeypatch.setattr(click, 'echo', lambda value=None, **_kwargs: emitted.append(value))

    OutputMixin.output(cli, itertools.chain(), SQLResult(image=b'png', image_protocol='iterm2'))

    assert logged == []
    assert emitted == ['\x1b]1337;File=name=Y2hhcnQucG5n;size=3;width=auto;height=auto;preserveAspectRatio=1;inline=1:cG5n\x07']


def test_output_emits_kitty_image_in_base64_chunks(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.explicit_pager = False
    cli.get_output_margin = lambda status=None: 1  # type: ignore[assignment]
    emitted: list[tuple[str | None, dict[str, Any]]] = []
    monkeypatch.setattr(click, 'echo', lambda value=None, **kwargs: emitted.append((value, kwargs)))
    image = b'x' * 4000

    OutputMixin.output(cli, itertools.chain(), SQLResult(image=image, image_protocol='kitty'))

    encoded = output_module.base64.b64encode(image).decode('ascii')
    assert emitted == [
        (f'\x1b_Ga=T,f=100,m=1;{encoded[:4096]}\x1b\\', {'nl': False}),
        (f'\x1b_Gm=0;{encoded[4096:]}\x1b\\', {'nl': False}),
        (None, {}),
    ]


@pytest.mark.parametrize('image_protocol', ['iterm2', 'kitty'])
def test_output_suppresses_images_in_windows_console(
    monkeypatch: pytest.MonkeyPatch,
    image_protocol: ImageProtocol,
) -> None:
    cli = make_bare_mycli()
    cli.prompt_session = cast(
        Any,
        SimpleNamespace(output=SimpleNamespace(get_size=lambda: SimpleNamespace(columns=80, rows=24))),
    )
    cli.get_output_margin = lambda status=None: 1  # type: ignore[assignment]
    emitted: list[str | None] = []
    monkeypatch.setattr(output_module, 'is_windows_console', lambda output: True)
    monkeypatch.setattr(click, 'echo', lambda value=None, **_kwargs: emitted.append(value))

    OutputMixin.output(cli, itertools.chain(), SQLResult(image=b'png', image_protocol=image_protocol))

    assert emitted == []


def test_is_windows_console_detects_prompt_toolkit_backends(monkeypatch: pytest.MonkeyPatch) -> None:
    class FakeWin32Output:
        pass

    class FakeWindows10Output:
        pass

    class FakeConEmuOutput:
        pass

    modules = (
        ('prompt_toolkit.output.win32', 'Win32Output', FakeWin32Output),
        ('prompt_toolkit.output.windows10', 'Windows10_Output', FakeWindows10Output),
        ('prompt_toolkit.output.conemu', 'ConEmuOutput', FakeConEmuOutput),
    )
    for module_name, class_name, output_class in modules:
        module = ModuleType(module_name)
        setattr(module, class_name, output_class)
        monkeypatch.setitem(sys.modules, module_name, module)
    monkeypatch.setattr(compat.sys, 'platform', 'win32')

    assert compat.is_windows_console(FakeWin32Output())
    assert compat.is_windows_console(FakeWindows10Output())
    assert compat.is_windows_console(FakeConEmuOutput())
    assert not compat.is_windows_console(object())


def test_is_windows_console_ignores_non_windows_platform(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(compat.sys, 'platform', 'cygwin')

    assert not compat.is_windows_console(object())


def test_get_output_margin_renders_prompt_once_and_counts_status_lines(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.prompt_lines = 0
    cli.prompt_format = 'ignored'
    cli.prompt_session = None
    cli.get_reserved_space = lambda: 2  # type: ignore[assignment]
    monkeypatch.setattr(output_module.repl_mode, 'render_prompt_string', lambda *_args: FormattedText([('', 'one\ntwo')]))
    monkeypatch.setattr(output_module.special, 'is_timing_enabled', lambda: True)

    margin = OutputMixin.get_output_margin(cli, 'ok\nwarning')

    assert margin == 7
    assert cli.prompt_lines == 2


def test_get_output_margin_uses_prompt_session_render_counter(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.prompt_lines = 0
    cli.prompt_session = cast(Any, SimpleNamespace(app=SimpleNamespace(render_counter=9)))
    cli.get_reserved_space = lambda: 1  # type: ignore[assignment]
    render_counters: list[int] = []

    def render_prompt_string(_cli: Any, _prompt_format: str, render_counter: int) -> FormattedText:
        render_counters.append(render_counter)
        return FormattedText([('', 'prompt')])

    monkeypatch.setattr(output_module.repl_mode, 'render_prompt_string', render_prompt_string)
    monkeypatch.setattr(output_module.special, 'is_timing_enabled', lambda: False)

    assert OutputMixin.get_output_margin(cli) == 2
    assert render_counters == [9]


def test_output_writes_lines_sinks_and_status(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.prompt_session = None
    cli.explicit_pager = False
    cli.get_output_margin = lambda status=None: 1  # type: ignore[assignment]
    logged: list[Any] = []
    tee: list[str] = []
    once: list[str] = []
    pipe_once: list[str] = []
    printed_lines: list[str] = []
    printed_status: list[Any] = []
    cli.log_output = lambda value: logged.append(value)  # type: ignore[assignment]
    monkeypatch.setattr(output_module.special, 'write_tee', lambda value: tee.append(value))
    monkeypatch.setattr(output_module.special, 'write_once', lambda value: once.append(value))
    monkeypatch.setattr(output_module.special, 'write_pipe_once', lambda value: pipe_once.append(value))
    monkeypatch.setattr(output_module.special, 'is_redirected', lambda: False)
    monkeypatch.setattr(output_module.special, 'is_pager_enabled', lambda: False)
    monkeypatch.setattr(click, 'secho', lambda value, **_kwargs: printed_lines.append(value))
    monkeypatch.setattr(prompt_toolkit, 'print_formatted_text', lambda text, style=None: printed_status.append(text))

    OutputMixin.output(cli, itertools.chain(['row 1', 'row 2']), SQLResult(status='done'))

    assert logged == ['row 1', 'row 2', 'done']
    assert tee == ['row 1', 'row 2']
    assert once == ['row 1', 'row 2']
    assert pipe_once == ['row 1', 'row 2']
    assert printed_lines == ['row 1', 'row 2']
    assert to_plain_text(printed_status[0]) == 'done'
    assert list(printed_status[0])[0][0].strip() == 'class:output.status'


def test_output_uses_prompt_session_size(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.prompt_session = cast(
        Any,
        SimpleNamespace(output=SimpleNamespace(get_size=lambda: SimpleNamespace(columns=80, rows=24))),
    )
    cli.explicit_pager = False
    cli.log_output = lambda value: None  # type: ignore[assignment]
    cli.get_output_margin = lambda status=None: 1  # type: ignore[assignment]
    printed_lines: list[str] = []
    monkeypatch.setattr(output_module.special, 'write_tee', lambda value: None)
    monkeypatch.setattr(output_module.special, 'write_once', lambda value: None)
    monkeypatch.setattr(output_module.special, 'write_pipe_once', lambda value: None)
    monkeypatch.setattr(output_module.special, 'is_redirected', lambda: False)
    monkeypatch.setattr(output_module.special, 'is_pager_enabled', lambda: False)
    monkeypatch.setattr(click, 'secho', lambda value, **_kwargs: printed_lines.append(value))

    OutputMixin.output(cli, itertools.chain(['row']), SQLResult())

    assert printed_lines == ['row']


def test_output_flushes_buffer_when_content_does_not_fit(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.prompt_session = None
    cli.explicit_pager = False
    cli.log_output = lambda value: None  # type: ignore[assignment]
    cli.get_output_margin = lambda status=None: output_module.DEFAULT_HEIGHT  # type: ignore[assignment]
    printed_lines: list[str] = []
    monkeypatch.setattr(output_module.special, 'write_tee', lambda value: None)
    monkeypatch.setattr(output_module.special, 'write_once', lambda value: None)
    monkeypatch.setattr(output_module.special, 'write_pipe_once', lambda value: None)
    monkeypatch.setattr(output_module.special, 'is_redirected', lambda: False)
    monkeypatch.setattr(output_module.special, 'is_pager_enabled', lambda: False)
    monkeypatch.setattr(click, 'secho', lambda value, **_kwargs: printed_lines.append(value))

    OutputMixin.output(cli, itertools.chain(['row 1', 'row 2']), SQLResult())

    assert printed_lines == ['row 1', 'row 2']


def test_output_switches_to_pager_when_content_does_not_fit(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.prompt_session = None
    cli.explicit_pager = False
    cli.log_output = lambda value: None  # type: ignore[assignment]
    cli.get_output_margin = lambda status=None: output_module.DEFAULT_HEIGHT  # type: ignore[assignment]
    paged_lines: list[str] = []
    monkeypatch.setattr(output_module.special, 'write_tee', lambda value: None)
    monkeypatch.setattr(output_module.special, 'write_once', lambda value: None)
    monkeypatch.setattr(output_module.special, 'write_pipe_once', lambda value: None)
    monkeypatch.setattr(output_module.special, 'is_redirected', lambda: False)
    monkeypatch.setattr(output_module.special, 'is_pager_enabled', lambda: True)
    monkeypatch.setattr(click, 'echo_via_pager', lambda values: paged_lines.extend(list(values)))

    OutputMixin.output(cli, itertools.chain(['row']), SQLResult())

    assert paged_lines == ['row\n']


def test_output_redirected_skips_screen_printing(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.prompt_session = None
    cli.log_output = lambda value: None  # type: ignore[assignment]
    cli.get_output_margin = lambda status=None: 1  # type: ignore[assignment]
    printed_lines: list[str] = []
    monkeypatch.setattr(output_module.special, 'write_tee', lambda value: None)
    monkeypatch.setattr(output_module.special, 'write_once', lambda value: None)
    monkeypatch.setattr(output_module.special, 'write_pipe_once', lambda value: None)
    monkeypatch.setattr(output_module.special, 'is_redirected', lambda: True)
    monkeypatch.setattr(output_module.special, 'is_pager_enabled', lambda: False)
    monkeypatch.setattr(click, 'secho', lambda value, **_kwargs: printed_lines.append(value))

    OutputMixin.output(cli, itertools.chain(['row']), SQLResult())

    assert printed_lines == []


def test_output_uses_warning_status_style(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.log_output = lambda value: None  # type: ignore[assignment]
    cli.get_output_margin = lambda status=None: 1  # type: ignore[assignment]
    printed_status: list[Any] = []
    monkeypatch.setattr(prompt_toolkit, 'print_formatted_text', lambda text, style=None: printed_status.append(text))

    OutputMixin.output(cli, itertools.chain([]), SQLResult(status='warning'), is_warnings_style=True)

    assert list(printed_status[0])[0][0].strip() == 'class:warnings.status'


def test_output_preserves_formatted_status(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.log_output = lambda value: None  # type: ignore[assignment]
    cli.get_output_margin = lambda status=None: 1  # type: ignore[assignment]
    printed_status: list[Any] = []
    monkeypatch.setattr(prompt_toolkit, 'print_formatted_text', lambda text, style=None: printed_status.append(text))
    status = FormattedText([('class:custom', 'formatted')])

    OutputMixin.output(cli, itertools.chain([]), SQLResult(status=status))

    assert to_plain_text(printed_status[0]) == 'formatted'


def test_output_sends_buffer_to_pager_when_pager_is_explicit(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.prompt_session = None
    cli.explicit_pager = True
    cli.log_output = lambda value: None  # type: ignore[assignment]
    cli.get_output_margin = lambda status=None: 1  # type: ignore[assignment]
    paged_lines: list[str] = []
    monkeypatch.setattr(output_module.special, 'write_tee', lambda value: None)
    monkeypatch.setattr(output_module.special, 'write_once', lambda value: None)
    monkeypatch.setattr(output_module.special, 'write_pipe_once', lambda value: None)
    monkeypatch.setattr(output_module.special, 'is_redirected', lambda: False)
    monkeypatch.setattr(output_module.special, 'is_pager_enabled', lambda: True)
    monkeypatch.setattr(click, 'echo_via_pager', lambda values: paged_lines.extend(list(values)))
    monkeypatch.setattr(prompt_toolkit, 'print_formatted_text', lambda text, style=None: None)

    OutputMixin.output(cli, itertools.chain(['row 1', 'row 2']), SQLResult())

    assert paged_lines == ['row 1\n', 'row 2\n']


def test_output_sends_buffer_to_explorer_and_clears_temporary_pager(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.prompt_session = None
    cli.explicit_pager = False
    cli.explorer_command = 'explorer-cmd'
    cli.explorer_exists = lambda: True  # type: ignore[method-assign]
    cli.log_output = lambda value: None  # type: ignore[assignment]
    cli.get_output_margin = lambda status=None: 1  # type: ignore[assignment]
    paged_lines: list[str] = []
    monkeypatch.delenv('PAGER', raising=False)
    monkeypatch.setattr(output_module.special, 'write_tee', lambda value: None)
    monkeypatch.setattr(output_module.special, 'write_once', lambda value: None)
    monkeypatch.setattr(output_module.special, 'write_pipe_once', lambda value: None)
    monkeypatch.setattr(output_module.special, 'is_redirected', lambda: False)
    monkeypatch.setattr(output_module.special, 'is_explorer_output', lambda: True)
    monkeypatch.setattr(click, 'echo_via_pager', lambda values: paged_lines.extend(list(values)))
    monkeypatch.setattr(prompt_toolkit, 'print_formatted_text', lambda text, style=None: None)

    OutputMixin.output(cli, itertools.chain(['row 1', 'row 2']), SQLResult())

    assert paged_lines == ['row 1\n', 'row 2\n']
    assert 'PAGER' not in output_module.os.environ


def test_output_sends_buffer_to_explorer_and_restores_existing_pager(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.prompt_session = None
    cli.explicit_pager = False
    cli.explorer_command = 'explorer-cmd'
    cli.explorer_exists = lambda: True  # type: ignore[method-assign]
    cli.log_output = lambda value: None  # type: ignore[assignment]
    cli.get_output_margin = lambda status=None: 1  # type: ignore[assignment]
    paged_lines: list[str] = []
    monkeypatch.setenv('PAGER', 'original-pager')
    monkeypatch.setattr(output_module.special, 'write_tee', lambda value: None)
    monkeypatch.setattr(output_module.special, 'write_once', lambda value: None)
    monkeypatch.setattr(output_module.special, 'write_pipe_once', lambda value: None)
    monkeypatch.setattr(output_module.special, 'is_redirected', lambda: False)
    monkeypatch.setattr(output_module.special, 'is_explorer_output', lambda: True)
    monkeypatch.setattr(click, 'echo_via_pager', lambda values: paged_lines.extend(list(values)))
    monkeypatch.setattr(prompt_toolkit, 'print_formatted_text', lambda text, style=None: None)

    OutputMixin.output(cli, itertools.chain(['row']), SQLResult())

    assert paged_lines == ['row\n']
    assert output_module.os.environ['PAGER'] == 'original-pager'


def test_output_reports_missing_configured_explorer(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.prompt_session = None
    cli.explicit_pager = False
    cli.explorer_command = 'missing-explorer'
    cli.explorer_exists = lambda: False  # type: ignore[method-assign]
    cli.log_output = lambda value: None  # type: ignore[assignment]
    cli.get_output_margin = lambda status=None: 1  # type: ignore[assignment]
    printed_lines: list[tuple[str, dict[str, Any]]] = []
    monkeypatch.setattr(output_module.special, 'write_tee', lambda value: None)
    monkeypatch.setattr(output_module.special, 'write_once', lambda value: None)
    monkeypatch.setattr(output_module.special, 'write_pipe_once', lambda value: None)
    monkeypatch.setattr(output_module.special, 'is_redirected', lambda: False)
    monkeypatch.setattr(output_module.special, 'is_explorer_output', lambda: True)
    monkeypatch.setattr(click, 'secho', lambda value, **kwargs: printed_lines.append((value, kwargs)))
    monkeypatch.setattr(prompt_toolkit, 'print_formatted_text', lambda text, style=None: None)

    OutputMixin.output(cli, itertools.chain(['row']), SQLResult())

    assert printed_lines == [('Configured explorer command not found: missing-explorer.', {'err': True, 'fg': 'red'})]


def test_configure_pager_uses_more_for_missing_less_on_windows(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.config = ConfigObj({'main': {'pager': 'less', 'enable_pager': 'True'}})
    pager_calls: list[str] = []
    monkeypatch.setattr(output_module, 'WIN', True)
    monkeypatch.setattr(output_module.shutil, 'which', lambda value: None)
    monkeypatch.setattr(output_module.special, 'set_pager', lambda value: pager_calls.append(value))
    monkeypatch.setattr(output_module.special, 'disable_pager', lambda: None)

    OutputMixin.configure_pager(cli)

    assert pager_calls == ['more']


def test_configure_pager_uses_myclirc_pager_and_sets_less(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.config = ConfigObj({'main': {'pager': 'config-pager', 'enable_pager': 'True'}})
    pager_calls: list[str] = []
    disabled: list[bool] = []
    monkeypatch.delenv('LESS', raising=False)
    monkeypatch.setattr(output_module.special, 'set_pager', lambda value: pager_calls.append(value))
    monkeypatch.setattr(output_module.special, 'disable_pager', lambda: disabled.append(True))

    OutputMixin.configure_pager(cli)

    assert pager_calls == ['config-pager']
    assert disabled == []
    assert cli.explicit_pager is True
    assert output_module.os.environ['LESS'] == '-RXF'


def test_configure_pager_disables_pager_when_configured(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.config = ConfigObj({'main': {'pager': '', 'enable_pager': 'False'}})
    pager_calls: list[str] = []
    disabled: list[bool] = []
    monkeypatch.setattr(output_module.special, 'set_pager', lambda value: pager_calls.append(value))
    monkeypatch.setattr(output_module.special, 'disable_pager', lambda: disabled.append(True))

    OutputMixin.configure_pager(cli)

    assert pager_calls == []
    assert disabled == [True]
    assert cli.explicit_pager is False


def test_explorer_exists_checks_command_name(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.explorer_command = 'explorer-cmd --with-args'
    checked: list[str] = []
    monkeypatch.setattr(output_module.shutil, 'which', lambda value: checked.append(value) or '/bin/explorer-cmd')  # type: ignore[func-returns-value]

    assert OutputMixin.explorer_exists(cli) is True
    assert checked == ['explorer-cmd']


def test_explorer_exists_returns_false_without_command() -> None:
    cli = make_bare_mycli()
    cli.explorer_command = ''

    assert OutputMixin.explorer_exists(cli) is False


def test_format_sqlresult_uses_redirect_formatter_and_appends_preamble_postamble() -> None:
    cli = make_bare_mycli()
    cli.main_formatter = DummyFormatter()
    cli.redirect_formatter = DummyFormatter()
    result = SQLResult(preamble='before', header=['id'], rows=[(1,)], postamble='after')

    formatted = list(OutputMixin.format_sqlresult(cli, result, is_redirected=True))

    assert formatted == ['before', 'plain output', 'after']
    assert cli.main_formatter.calls == []
    assert cli.redirect_formatter.calls


def test_format_sqlresult_uses_explorer_formatter_and_trims_footer(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.main_formatter = DummyFormatter()
    cli.explorer_formatter = DummyFormatter()
    cli.explorer_trim_footer = True
    monkeypatch.setattr(output_module.special, 'is_explorer_output', lambda: True)
    cli.explorer_formatter.format_output = lambda *args, **kwargs: ['header', 'row', 'footer']  # type: ignore[method-assign]
    result = SQLResult(header=['id'], rows=[(1,)])

    formatted = list(OutputMixin.format_sqlresult(cli, result))

    assert formatted == ['header', 'row']
    assert cli.main_formatter.calls == []


def test_format_sqlresult_uses_null_string_when_default_missing_value_is_configured() -> None:
    cli = make_bare_mycli()
    cli.main_formatter = DummyFormatter()
    result = SQLResult(header=['id'], rows=[(None,)])

    list(OutputMixin.format_sqlresult(cli, result, null_string='<null>'))

    _, kwargs = cli.main_formatter.calls[-1]
    assert kwargs['missing_value'] == '<null>'


def test_format_sqlresult_for_cursor_sets_column_types_and_alignment(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.main_formatter = DummyFormatter()
    monkeypatch.setattr(output_module, 'Cursor', FakeCursorBase)
    rows = FakeCursorBase(rows=[(1, 'name')], rowcount=1, description=[('id', 3), ('name', 253)])
    result = SQLResult(header=['id', 'name'], rows=cast(Any, rows))

    assert list(OutputMixin.format_sqlresult(cli, result, numeric_alignment='left')) == ['plain output']

    _, kwargs = cli.main_formatter.calls[-1]
    assert kwargs['column_types'] == [int, str]
    assert kwargs['colalign'] == ['left', 'left']


def test_format_sqlresult_for_empty_cursor_uses_empty_column_metadata(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.main_formatter = DummyFormatter()
    monkeypatch.setattr(output_module, 'Cursor', FakeCursorBase)
    rows = FakeCursorBase(rowcount=0, description=[('id', 3)])
    result = SQLResult(header=['id'], rows=cast(Any, rows))

    list(OutputMixin.format_sqlresult(cli, result))

    _, kwargs = cli.main_formatter.calls[-1]
    assert kwargs['column_types'] == []
    assert kwargs['colalign'] == []


def test_format_sqlresult_materializes_cursor_rows_when_width_is_limited(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.main_formatter = DummyFormatter()
    monkeypatch.setattr(output_module, 'Cursor', FakeCursorBase)
    rows = FakeCursorBase(rows=[(1,)], rowcount=1, description=[('id', 3)])
    result = SQLResult(header=['id'], rows=cast(Any, rows))

    list(OutputMixin.format_sqlresult(cli, result, max_width=100))

    formatted_rows = cli.main_formatter.calls[-1][0][0]
    assert formatted_rows == [(1,)]


def test_format_sqlresult_splits_string_formatter_output() -> None:
    cli = make_bare_mycli()
    cli.main_formatter = DummyFormatter()
    cli.main_formatter.format_output = lambda *args, **kwargs: 'one\ntwo'  # type: ignore[method-assign]
    result = SQLResult(header=['id'], rows=[(1,)])

    assert list(OutputMixin.format_sqlresult(cli, result)) == ['one', 'two']


def test_format_sqlresult_switches_to_vertical_when_first_line_is_too_wide() -> None:
    cli = make_bare_mycli()
    cli.main_formatter = DummyFormatter()
    result = SQLResult(header=['id'], rows=[(1,)])

    assert list(OutputMixin.format_sqlresult(cli, result, max_width=2)) == ['vertical output']


def test_format_sqlresult_splits_string_vertical_output_when_table_is_too_wide() -> None:
    cli = make_bare_mycli()
    cli.main_formatter = DummyFormatter()

    def format_output(rows, header, format_name=None, **kwargs):
        if format_name == 'vertical':
            return 'vertical one\nvertical two'
        return ['too wide']

    cli.main_formatter.format_output = format_output  # type: ignore[method-assign]
    result = SQLResult(header=['id'], rows=[(1,)])

    assert list(OutputMixin.format_sqlresult(cli, result, max_width=2)) == ['vertical one', 'vertical two']


def test_format_sqlresult_keeps_table_when_first_line_fits_width() -> None:
    cli = make_bare_mycli()
    cli.main_formatter = DummyFormatter()
    result = SQLResult(header=['id'], rows=[(1,)])

    assert list(OutputMixin.format_sqlresult(cli, result, max_width=100)) == ['plain output']


def test_get_reserved_space_caps_ratio(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    monkeypatch.setattr(shutil, 'get_terminal_size', lambda *args, **kwargs: (120, 40))

    assert OutputMixin.get_reserved_space(cli) == 8



from __future__ import annotations

from io import TextIOWrapper
from pathlib import Path
from types import SimpleNamespace
from typing import Any, cast

import pytest

import mycli.client as client_module
from mycli.client import MyCli


def write_myclirc(tmp_path: Path, content: str) -> str:
    myclirc = tmp_path / 'myclirc'
    myclirc.write_text(content, encoding='utf-8')
    return str(myclirc)


def patch_constructor_side_effects(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(MyCli, 'system_config_files', [])
    monkeypatch.setattr(MyCli, 'initialize_logging', lambda self: None)
    monkeypatch.setattr(MyCli, 'register_special_commands', lambda self: None)
    monkeypatch.setattr(client_module, 'get_mylogin_cnf_path', lambda: None)


def test_init_reports_invalid_ssl_mode(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    patch_constructor_side_effects(monkeypatch)
    echo_calls: list[tuple[str, dict[str, Any]]] = []
    monkeypatch.setattr(MyCli, 'echo', lambda self, message, **kwargs: echo_calls.append((message, kwargs)))
    myclirc = write_myclirc(
        tmp_path,
        """
        [connection]
        default_ssl_mode = invalid
        """,
    )

    cli = MyCli(myclirc=myclirc)

    assert cli.ssl_mode is None
    assert echo_calls == [('Invalid config option provided for ssl_mode (invalid); ignoring.', {'err': True, 'fg': 'red'})]


def test_init_configures_image_protocol(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    patch_constructor_side_effects(monkeypatch)
    myclirc = write_myclirc(
        tmp_path,
        """
        [dataframe]
        image_protocol = iterm2
        plot_scale_factor = 1.5
        plot_ppi = 144
        plot_theme = dark
        """,
    )

    cli = MyCli(myclirc=myclirc)

    assert cli.image_protocol == 'iterm2'
    assert cli.plot_scale_factor == 1.5
    assert cli.plot_ppi == 144
    assert isinstance(cli.plot_ppi, int)
    assert cli.plot_theme == 'dark'


def test_init_uses_default_plot_theme_for_empty_value(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    patch_constructor_side_effects(monkeypatch)
    myclirc = write_myclirc(
        tmp_path,
        """
        [dataframe]
        plot_theme =
        """,
    )

    cli = MyCli(myclirc=myclirc)

    assert cli.plot_theme == 'carbong90'


def test_init_configures_indexed_column_suffix(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    patch_constructor_side_effects(monkeypatch)
    myclirc = write_myclirc(
        tmp_path,
        """
        [main]
        indexed_column_suffix = " [indexed]"
        """,
    )

    cli = MyCli(myclirc=myclirc)

    assert cli.completer.indexed_column_suffix == ' [indexed]'


def test_init_allows_empty_indexed_column_suffix(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    patch_constructor_side_effects(monkeypatch)
    myclirc = write_myclirc(
        tmp_path,
        """
        [main]
        indexed_column_suffix =
        """,
    )

    cli = MyCli(myclirc=myclirc)

    assert cli.completer.indexed_column_suffix == ''


def test_init_configures_kitty_image_protocol(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    patch_constructor_side_effects(monkeypatch)
    myclirc = write_myclirc(
        tmp_path,
        """
        [dataframe]
        image_protocol = kitty
        """,
    )

    cli = MyCli(myclirc=myclirc)

    assert cli.image_protocol == 'kitty'


def test_init_reports_invalid_image_protocol(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    patch_constructor_side_effects(monkeypatch)
    echo_calls: list[tuple[str, dict[str, Any]]] = []
    monkeypatch.setattr(MyCli, 'echo', lambda self, message, **kwargs: echo_calls.append((message, kwargs)))
    myclirc = write_myclirc(
        tmp_path,
        """
        [dataframe]
        image_protocol = sixel
        """,
    )

    cli = MyCli(myclirc=myclirc)

    assert cli.image_protocol == 'none'
    assert echo_calls == [('Invalid config option provided for image_protocol (sixel); disabling.', {'err': True, 'fg': 'red'})]


def test_init_honors_explicit_show_warnings(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    patch_constructor_side_effects(monkeypatch)
    show_warnings_calls: list[bool] = []
    monkeypatch.setattr(client_module.special, 'set_show_warnings_enabled', lambda value: show_warnings_calls.append(value))
    myclirc = write_myclirc(tmp_path, '')

    MyCli(myclirc=myclirc, show_warnings=True)

    assert show_warnings_calls == [True]


def test_init_uses_cli_verbosity_override(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    patch_constructor_side_effects(monkeypatch)
    myclirc = write_myclirc(tmp_path, '')

    cli = MyCli(myclirc=myclirc, cli_verbosity=2)

    assert cli.verbosity == 2


def test_init_uses_existing_xdg_config_when_myclirc_is_not_given(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    patch_constructor_side_effects(monkeypatch)
    xdg_config = tmp_path / 'xdg' / 'mycli' / 'myclirc'
    xdg_config.parent.mkdir(parents=True)
    xdg_config.write_text('', encoding='utf-8')
    monkeypatch.setattr(MyCli, 'xdg_config_file', str(xdg_config))

    cli = MyCli(myclirc=None)

    assert cli.config.filename == str(xdg_config)


def test_init_uses_default_myclirc_when_xdg_config_is_missing(monkeypatch: pytest.MonkeyPatch) -> None:
    patch_constructor_side_effects(monkeypatch)
    config_file_args: list[list[str | Any]] = []
    monkeypatch.setattr(MyCli, 'xdg_config_file', '/missing/xdg/myclirc')
    monkeypatch.setattr(client_module.os.path, 'exists', lambda path: False)
    monkeypatch.setattr(client_module, 'write_default_config', lambda destination: None)

    original_read_config_files = client_module.read_config_files

    def read_config_files(files: list[str | Any], *args: Any, **kwargs: Any) -> Any:
        config_file_args.append(files)
        return original_read_config_files(files, *args, **kwargs)

    monkeypatch.setattr(client_module, 'read_config_files', read_config_files)

    MyCli(myclirc=None)

    assert config_file_args[0] == ['~/.myclirc']


def test_init_writes_default_config_when_user_config_is_missing(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    patch_constructor_side_effects(monkeypatch)
    write_calls: list[str] = []
    myclirc = tmp_path / 'missing-myclirc'
    monkeypatch.setattr(client_module, 'write_default_config', lambda destination: write_calls.append(destination))

    MyCli(myclirc=str(myclirc))

    assert write_calls == [str(myclirc)]


def test_init_opens_audit_log_from_config(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    patch_constructor_side_effects(monkeypatch)
    audit_log = tmp_path / 'audit.log'
    myclirc = write_myclirc(
        tmp_path,
        f"""
        [main]
        audit_log = {audit_log}
        """,
    )

    cli = MyCli(myclirc=myclirc)
    try:
        assert isinstance(cli.logfile, TextIOWrapper)
        assert Path(cli.logfile.name) == audit_log
    finally:
        if isinstance(cli.logfile, TextIOWrapper):
            cli.logfile.close()


def test_init_disables_audit_log_when_file_cannot_be_opened(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    patch_constructor_side_effects(monkeypatch)
    echo_calls: list[tuple[str, dict[str, Any]]] = []
    monkeypatch.setattr(MyCli, 'echo', lambda self, message, **kwargs: echo_calls.append((message, kwargs)))
    myclirc = write_myclirc(
        tmp_path,
        f"""
        [main]
        audit_log = {tmp_path}
        """,
    )

    cli = MyCli(myclirc=myclirc)

    assert cli.logfile is False
    assert echo_calls == [
        (
            'Error: Unable to open the audit log file. Your queries will not be logged.',
            {'err': True, 'fg': 'red'},
        )
    ]


def test_init_reports_unreadable_mylogin_cnf(monkeypatch: pytest.MonkeyPatch, tmp_path: Path, capsys: pytest.CaptureFixture[str]) -> None:
    patch_constructor_side_effects(monkeypatch)
    mylogin_path = tmp_path / 'mylogin.cnf'
    mylogin_path.write_text('bad', encoding='utf-8')
    monkeypatch.setattr(client_module, 'get_mylogin_cnf_path', lambda: str(mylogin_path))
    monkeypatch.setattr(client_module, 'open_mylogin_cnf', lambda path: None)
    myclirc = write_myclirc(tmp_path, '')

    MyCli(myclirc=myclirc)

    assert 'Error: Unable to read login path file.' in capsys.readouterr().out


def test_close_stops_schema_prefetcher_and_closes_sqlexecute() -> None:
    cli = MyCli.__new__(MyCli)
    stopped: list[bool] = []
    closed: list[bool] = []
    tunnel_closed: list[bool] = []
    cli.schema_prefetcher = SimpleNamespace(stop=lambda: stopped.append(True))
    cli.sqlexecute = SimpleNamespace(close=lambda: closed.append(True))  # type: ignore[assignment]
    cast(Any, cli).ssh_tunnel = SimpleNamespace(close=lambda: tunnel_closed.append(True))

    MyCli.close(cli)

    assert stopped == [True]
    assert closed == [True]
    assert tunnel_closed == [True]


def test_close_swallows_cleanup_errors() -> None:
    cli = MyCli.__new__(MyCli)

    def fail() -> None:
        raise RuntimeError('cleanup failed')

    cli.schema_prefetcher = SimpleNamespace(stop=fail)
    cli.sqlexecute = SimpleNamespace(close=fail)  # type: ignore[assignment]
    cast(Any, cli).ssh_tunnel = SimpleNamespace(close=fail)

    MyCli.close(cli)


def test_invalidate_prompt_session_invalidates_prompt_app() -> None:
    cli = MyCli.__new__(MyCli)
    invalidate_calls: list[bool] = []
    cast(Any, cli).prompt_session = SimpleNamespace(app=SimpleNamespace(invalidate=lambda: invalidate_calls.append(True)))

    MyCli._invalidate_prompt_session(cli)

    assert invalidate_calls == [True]


def test_run_cli_delegates_to_main_repl(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = MyCli.__new__(MyCli)
    calls: list[MyCli] = []
    monkeypatch.setattr(client_module.repl_package, 'main_repl', lambda target: calls.append(target))

    MyCli.run_cli(cli)

    assert calls == [cli]



from __future__ import annotations

import pytest

from mycli.password_sources import (
    KNOWN_PASSWORD_SOURCES,
    PasswordCandidates,
)


def test_resolve_uses_fixed_precedence_instead_of_registration_order() -> None:
    candidates = PasswordCandidates()
    candidates.add_value('keyring', 'keyring-secret')
    candidates.add_value('dsn', 'dsn-secret')
    candidates.add_value('literal', 'cli-secret')

    selected = candidates.resolve(KNOWN_PASSWORD_SOURCES)

    assert selected is not None
    assert selected.source == 'literal'
    assert selected.value == 'cli-secret'


def test_resolve_treats_empty_string_as_password() -> None:
    candidates = PasswordCandidates()
    candidates.add_value('literal', '')
    candidates.add_value('file', 'file-secret')

    selected = candidates.resolve(KNOWN_PASSWORD_SOURCES)

    assert selected is not None
    assert selected.source == 'literal'
    assert selected.value == ''


def test_resolve_skips_lower_priority_lazy_loader() -> None:
    loader_calls: list[str] = []

    def load_vault_password() -> str:
        loader_calls.append('vault')
        return 'vault-secret'

    candidates = PasswordCandidates()
    candidates.add_loader('vault', load_vault_password)
    candidates.add_value('environment', 'environment-secret')

    selected = candidates.resolve(KNOWN_PASSWORD_SOURCES)

    assert selected is not None
    assert selected.source == 'environment'
    assert loader_calls == []


def test_resolve_falls_back_when_a_loader_returns_none() -> None:
    candidates = PasswordCandidates()
    candidates.add_loader('vault', lambda: None)
    candidates.add_value('login_path', 'mylogin-secret')

    selected = candidates.resolve(KNOWN_PASSWORD_SOURCES)

    assert selected is not None
    assert selected.source == 'login_path'
    assert selected.value == 'mylogin-secret'


def test_resolve_fallback_loader() -> None:
    candidates = PasswordCandidates()
    candidates.add_loader('fallback', lambda: 'fallback-secret')

    selected = candidates.resolve(['fallback'])

    assert selected is not None
    assert selected.source == 'fallback'
    assert selected.value == 'fallback-secret'


def test_resolve_skips_unknown_source_and_returns_none(capsys: pytest.CaptureFixture[str]) -> None:
    candidates = PasswordCandidates()

    assert candidates.resolve(['unknown']) is None
    assert capsys.readouterr().err == 'Skipping unknown password source: unknown.\n'



from __future__ import annotations

from dataclasses import dataclass
import os
import shlex
import socket
import subprocess
import tempfile
import threading
import time
from typing import Literal

from mycli.compat import WIN

DEFAULT_LOCALHOST = 'localhost'
DEFAULT_SSH_EXECUTABLE = 'ssh'
DEFAULT_TUNNEL_METHOD: Literal['auto', 'socket', 'port'] = 'auto'
MAX_UNIX_SOCKET_PATH_BYTES = 103


class SshTunnelError(RuntimeError):
    pass


def _find_free_local_port(local_host: str | None) -> int:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.bind((local_host or DEFAULT_LOCALHOST, 0))
        return int(sock.getsockname()[1])


def _make_local_socket_path() -> str:
    fd, path = tempfile.mkstemp(prefix='mycli-ssh-', suffix='.sock')
    os.close(fd)
    os.unlink(path)
    _check_local_socket_path_limit(path)
    return path


def _check_local_socket_path_limit(path: str) -> None:
    path_bytes = len(os.fsencode(path))
    if path_bytes > MAX_UNIX_SOCKET_PATH_BYTES:
        raise SshTunnelError(
            f'Local SSH socket path is too long for sockaddr_un: {path_bytes} bytes > {MAX_UNIX_SOCKET_PATH_BYTES} bytes: {path}'
        )


@dataclass(slots=True)
class SshTunnelTarget:
    ssh_target: str
    ssh_port: int | None = None

    @classmethod
    def parse(cls, value: str) -> SshTunnelTarget:
        target, separator, port = value.rpartition(':')
        if separator and port.isdigit() and ']' not in port:
            return cls(target, int(port))
        return cls(value, None)


class SshTunnel:
    def __init__(
        self,
        *,
        ssh_target: str,
        remote_host: str,
        remote_port: int,
        remote_socket: str | None = None,
        ssh_executable: str = DEFAULT_SSH_EXECUTABLE,
        ssh_config_options: str | None = None,
        ssh_cli_options: str | None = None,
        ssh_port: int | None = None,
        ready_timeout: float = 30.0,
        tunnel_method: Literal['auto', 'socket', 'port'] = DEFAULT_TUNNEL_METHOD,
    ) -> None:
        self.ssh_executable = ssh_executable
        self.ssh_config_options = ssh_config_options
        self.ssh_cli_options = ssh_cli_options
        self.ssh_target = ssh_target
        self.ssh_port = ssh_port
        self.remote_host = remote_host
        self.remote_port = remote_port
        self.remote_socket = remote_socket
        self.tunnel_method = tunnel_method
        if tunnel_method == 'auto':
            self.true_tunnel_method: Literal['socket', 'port'] = 'port' if WIN else 'socket'
        else:
            self.true_tunnel_method = tunnel_method
        self.local_socket = _make_local_socket_path() if self.true_tunnel_method == 'socket' else None
        self.local_host = DEFAULT_LOCALHOST if self.true_tunnel_method == 'port' else None
        self.local_port = _find_free_local_port(self.local_host) if self.true_tunnel_method == 'port' else None
        self.ready_timeout = ready_timeout
        self.process: subprocess.Popen | None = None
        self._startup_error: OSError | None = None
        self._ready = threading.Event()
        self._failed = threading.Event()
        self._thread: threading.Thread | None = None

    @classmethod
    def from_target(
        cls,
        ssh_jump_spec: str,
        *,
        remote_host: str,
        remote_port: int,
        remote_socket: str | None = None,
        ssh_executable: str = DEFAULT_SSH_EXECUTABLE,
        ssh_config_options: str | None = None,
        ssh_cli_options: str | None = None,
        tunnel_method: Literal['auto', 'socket', 'port'] = DEFAULT_TUNNEL_METHOD,
    ) -> SshTunnel:
        target = SshTunnelTarget.parse(ssh_jump_spec)
        return cls(
            ssh_target=target.ssh_target,
            ssh_port=target.ssh_port,
            remote_host=remote_host,
            remote_port=remote_port,
            remote_socket=remote_socket,
            ssh_executable=ssh_executable,
            ssh_config_options=ssh_config_options,
            ssh_cli_options=ssh_cli_options,
            tunnel_method=tunnel_method,
        )

    def _forward_spec(self) -> str:
        if self.local_socket:
            if self.remote_socket:
                return f'{self.local_socket}:{self.remote_socket}'
            return f'{self.local_socket}:{self.remote_host}:{self.remote_port}'
        else:
            if self.remote_socket:
                return f'{self.local_host}:{self.local_port}:{self.remote_socket}'
            return f'{self.local_host}:{self.local_port}:{self.remote_host}:{self.remote_port}'

    def command(self) -> list[str]:
        opts = shlex.split(self.ssh_config_options or '')
        opts.extend(shlex.split(self.ssh_cli_options or ''))
        command = [
            self.ssh_executable,
            *opts,
            '-N',
            '-L',
            self._forward_spec(),
        ]
        if self.ssh_port is not None:
            command.extend(['-p', str(self.ssh_port)])
        command.append(self.ssh_target)
        return command

    def start(self) -> None:
        self._thread = threading.Thread(target=self._run, name='mycli-ssh-tunnel', daemon=True)
        self._thread.start()
        deadline = time.monotonic() + self.ready_timeout
        while time.monotonic() < deadline:
            if self._failed.is_set():
                self.close()
                if self._startup_error is not None:
                    raise SshTunnelError(f'Unable to start SSH tunnel process: {self._startup_error}') from self._startup_error
                raise SshTunnelError('SSH tunnel process exited before it was ready.')
            if self._is_listening():
                self._ready.set()
                return
            time.sleep(0.05)
        self.close()
        raise SshTunnelError('Timed out waiting for SSH tunnel to become ready.')

    def close(self) -> None:
        process = self.process
        if process is not None and process.poll() is None:
            process.terminate()
            try:
                process.wait(timeout=5)
            except subprocess.TimeoutExpired:
                process.kill()
                process.wait()
        if self._thread is not None and self._thread.is_alive():
            self._thread.join(timeout=5)
        try:
            if self.local_socket:
                os.unlink(self.local_socket)
        except FileNotFoundError:
            pass

    def _run(self) -> None:
        try:
            self.process = subprocess.Popen(
                self.command(),
                stdin=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
        except OSError as exc:
            self._startup_error = exc
            self._failed.set()
            return
        return_code = self.process.wait()
        if return_code != 0 and not self._ready.is_set():
            self._failed.set()

    def _is_listening(self) -> bool:
        try:
            if self.local_socket:
                with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as sock:
                    sock.settimeout(0.05)
                    sock.connect(self.local_socket)
                    return True
            elif self.local_host and self.local_port:
                with socket.create_connection((self.local_host, self.local_port), timeout=0.05):
                    return True
            else:
                return False
        except OSError:
            return False



from collections.abc import Mapping

from mycli.packages.special.dsn_aliases import DsnAliases


class DummyConfig(dict):
    def __init__(self, initial: Mapping[str, object] | None = None) -> None:
        super().__init__(initial or {})
        self.encoding: str | None = None
        self.write_calls = 0

    def write(self) -> None:
        self.write_calls += 1


def test_from_config_returns_instance_with_same_config() -> None:
    config = DummyConfig()

    aliases = DsnAliases.from_config(config)

    assert isinstance(aliases, DsnAliases)
    assert aliases.config is config


def test_list_and_get_use_alias_dsn_section() -> None:
    config = DummyConfig({
        'alias_dsn': {
            'prod': 'mysql://prod/db',
            'staging': 'mysql://staging/db',
        },
    })
    aliases = DsnAliases(config)

    assert aliases.list() == ['prod', 'staging']
    assert aliases.get('prod') == 'mysql://prod/db'
    assert aliases.get('missing') is None


def test_list_returns_empty_list_when_section_is_missing() -> None:
    aliases = DsnAliases(DummyConfig())

    assert aliases.list() == []


def test_save_creates_section_sets_encoding_and_writes_config() -> None:
    config = DummyConfig()
    aliases = DsnAliases(config)

    result = aliases.save('prod', 'mysql://prod/db')

    assert result == 'Saved: prod'
    assert config.encoding == 'utf-8'
    assert config == {'alias_dsn': {'prod': 'mysql://prod/db'}}
    assert config.write_calls == 1


def test_save_updates_existing_section_and_writes_config() -> None:
    config = DummyConfig({'alias_dsn': {'prod': 'mysql://prod/db'}})
    aliases = DsnAliases(config)

    result = aliases.save('staging', 'mysql://staging/db')

    assert result == 'Saved: staging'
    assert config.encoding == 'utf-8'
    assert config['alias_dsn'] == {
        'prod': 'mysql://prod/db',
        'staging': 'mysql://staging/db',
    }
    assert config.write_calls == 1


def test_delete_removes_existing_alias_and_writes_config() -> None:
    config = DummyConfig({'alias_dsn': {'prod': 'mysql://prod/db'}})
    aliases = DsnAliases(config)

    result = aliases.delete('prod')

    assert result == 'Deleted: prod'
    assert config['alias_dsn'] == {}
    assert config.write_calls == 1


def test_delete_returns_not_found_without_writing_config() -> None:
    config = DummyConfig({'alias_dsn': {'prod': 'mysql://prod/db'}})
    aliases = DsnAliases(config)

    result = aliases.delete('missing')

    assert result == 'Not Found: missing'
    assert config['alias_dsn'] == {'prod': 'mysql://prod/db'}
    assert config.write_calls == 0


def test_delete_returns_not_found_when_section_is_missing() -> None:
    config = DummyConfig()
    aliases = DsnAliases(config)

    result = aliases.delete('missing')

    assert result == 'Not Found: missing'
    assert config == {}
    assert config.write_calls == 0



from __future__ import annotations

import logging
from pathlib import Path
from typing import Any

import pytest

from mycli import client_commands
from mycli.client_commands import ClientCommandsMixin
from mycli.packages.sqlresult import SQLResult


class DummyClient(ClientCommandsMixin):
    def __init__(self) -> None:
        self.echo_calls: list[tuple[tuple[Any, ...], dict[str, Any]]] = []

    def echo(self, *args: Any, **kwargs: Any) -> None:
        self.echo_calls.append((args, kwargs))

    def reconnect(self, database: str = '') -> bool:
        self.reconnect_database = database
        return True

    def refresh_completions(self, reset: bool = False) -> list[SQLResult]:
        return [SQLResult(status=f'refresh {reset}')]


class FakeFormatter:
    def __init__(self, *, supported_formats: list[str] | None = None, fail: bool = False) -> None:
        self.supported_formats = supported_formats or ['ascii', 'csv']
        self.fail = fail
        self.values: list[str] = []

    @property
    def format_name(self) -> str:
        return self.values[-1]

    @format_name.setter
    def format_name(self, value: str) -> None:
        if self.fail:
            raise ValueError
        self.values.append(value)


class FakeSQLExecute:
    def __init__(self, *, dbname: str = 'old_db', user: str = 'alice') -> None:
        self.dbname = dbname
        self.user = user
        self.changed_to: list[str] = []
        self.runs: list[str] = []

    def change_db(self, dbname: str) -> None:
        self.changed_to.append(dbname)
        self.dbname = dbname

    def run(self, query: str) -> list[SQLResult]:
        self.runs.append(query)
        return [SQLResult(status=f'ran {query}')]


@pytest.fixture(autouse=True)
def patch_sql_execute(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(client_commands, 'SQLExecute', FakeSQLExecute)


def result_statuses(results: Any) -> list[str | None]:
    return [result.status for result in list(results)]


def test_register_special_commands_registers_expected_commands(monkeypatch: pytest.MonkeyPatch) -> None:
    client = DummyClient()
    calls: list[tuple[Any, ...]] = []
    monkeypatch.setattr(client_commands.special, 'register_special_command', lambda *args, **kwargs: calls.append((*args, kwargs)))

    client.register_special_commands()

    assert [call[1] for call in calls] == ['use', 'connect', 'rehash', 'tableformat', 'redirectformat', 'source', 'prompt']
    assert calls[0][0] == client.change_db
    assert calls[1][0] == client.manual_reconnect
    assert calls[2][0] == client.refresh_completions
    assert calls[3][0] == client.change_table_format
    assert calls[4][0] == client.change_redirect_format
    assert calls[5][0] == client.execute_from_file
    assert calls[6][0] == client.change_prompt_format


def test_manual_reconnect_reports_not_connected() -> None:
    client = DummyClient()

    def fake_reconnect(database: str = '') -> bool:
        client.reconnect_database = database
        return False

    client.reconnect = fake_reconnect  # type: ignore[method-assign]

    assert result_statuses(client.manual_reconnect('new_db')) == ['Not connected']
    assert client.reconnect_database == 'new_db'


def test_manual_reconnect_without_database_returns_empty_result() -> None:
    client = DummyClient()

    assert list(client.manual_reconnect()) == [SQLResult()]
    assert client.reconnect_database == ''


def test_manual_reconnect_with_database_delegates_to_change_db(monkeypatch: pytest.MonkeyPatch) -> None:
    client = DummyClient()
    changed: list[str] = []

    def fake_change_db(arg: str, **_: Any) -> Any:
        changed.append(arg)
        yield SQLResult(status='changed')

    monkeypatch.setattr(client, 'change_db', fake_change_db)

    assert result_statuses(client.manual_reconnect('new_db')) == ['changed']
    assert changed == ['new_db']


def test_change_table_format_reports_supported_formats_on_error() -> None:
    client = DummyClient()
    client.main_formatter = FakeFormatter(supported_formats=['plain', 'csv'], fail=True)

    assert result_statuses(client.change_table_format('bad')) == ['Table format bad not recognized. Allowed formats:\n\tplain\n\tcsv']


def test_change_table_format_updates_formatter() -> None:
    client = DummyClient()
    client.main_formatter = FakeFormatter()

    assert result_statuses(client.change_table_format('csv')) == ['Changed table format to csv']
    assert client.main_formatter.values == ['csv']


def test_change_redirect_format_updates_formatter() -> None:
    client = DummyClient()
    client.redirect_formatter = FakeFormatter()

    assert result_statuses(client.change_redirect_format('csv')) == ['Changed redirect format to csv']
    assert client.redirect_formatter.values == ['csv']


def test_change_redirect_format_reports_supported_formats_on_error() -> None:
    client = DummyClient()
    client.redirect_formatter = FakeFormatter(supported_formats=['plain', 'json'], fail=True)

    assert result_statuses(client.change_redirect_format('bad')) == [
        'Redirect format bad not recognized. Allowed formats:\n\tplain\n\tjson'
    ]


def test_change_db_unquotes_mysql_identifier(monkeypatch: pytest.MonkeyPatch) -> None:
    client = DummyClient()
    client.sqlexecute = FakeSQLExecute()
    title_calls: list[DummyClient] = []
    monkeypatch.setattr(client_commands, 'set_all_external_titles', lambda value: title_calls.append(value))

    assert result_statuses(client.change_db('`new``db`')) == ['You are now connected to database "new`db" as user "alice"']
    assert client.sqlexecute.changed_to == ['new`db']
    assert title_calls == [client]


def test_change_db_reports_when_database_is_already_selected(monkeypatch: pytest.MonkeyPatch) -> None:
    client = DummyClient()
    client.sqlexecute = FakeSQLExecute(dbname='same_db')
    title_calls: list[DummyClient] = []
    monkeypatch.setattr(client_commands, 'set_all_external_titles', lambda value: title_calls.append(value))

    assert result_statuses(client.change_db('same_db')) == ['You are already connected to database "same_db" as user "alice"']
    assert client.sqlexecute.changed_to == []
    assert title_calls == [client]


def test_change_db_without_argument_reports_error(monkeypatch: pytest.MonkeyPatch) -> None:
    client = DummyClient()
    secho_calls: list[tuple[str, dict[str, Any]]] = []
    monkeypatch.setattr(client_commands.click, 'secho', lambda message, **kwargs: secho_calls.append((message, kwargs)))

    assert list(client.change_db('')) == []
    assert secho_calls == [('No database selected', {'err': True, 'fg': 'red'})]


def test_execute_from_file_requires_filename() -> None:
    client = DummyClient()

    assert list(client.execute_from_file('')) == [SQLResult(status='Missing required argument: filename.')]


def test_execute_from_file_reports_open_errors() -> None:
    client = DummyClient()

    result = list(client.execute_from_file('/does/not/exist.sql'))

    assert len(result) == 1
    assert result[0].status is not None
    assert '/does/not/exist.sql' in result[0].status


def test_execute_from_file_stops_when_destructive_query_is_rejected(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    client = DummyClient()
    sql_file = tmp_path / 'query.sql'
    sql_file.write_text('drop table users;', encoding='utf-8')
    client.destructive_warning = True
    client.destructive_keywords = {'drop'}
    monkeypatch.setattr(client_commands, 'confirm_destructive_query', lambda keywords, query: False)

    assert list(client.execute_from_file(str(sql_file))) == [SQLResult(status='Wise choice. Command execution stopped.')]


def test_execute_from_file_runs_file_query(tmp_path: Path) -> None:
    client = DummyClient()
    sql_file = tmp_path / 'query.sql'
    sql_file.write_text('select 1;', encoding='utf-8')
    client.destructive_warning = False
    client.destructive_keywords = set()
    client.sqlexecute = FakeSQLExecute()

    assert list(client.execute_from_file(str(sql_file))) == [SQLResult(status='ran select 1;')]
    assert client.sqlexecute.runs == ['select 1;']


def test_change_prompt_format_requires_argument() -> None:
    client = DummyClient()

    assert client.change_prompt_format('') == [SQLResult(status='Missing required argument, format.')]


def test_change_prompt_format_updates_prompt_format() -> None:
    client = DummyClient()

    assert client.change_prompt_format('\\u> ') == [SQLResult(status='Changed prompt format to \\u> ')]
    assert client.prompt_format == '\\u> '


def test_initialize_logging_uses_null_handler_for_none_level(monkeypatch: pytest.MonkeyPatch) -> None:
    client = DummyClient()
    client.config = {'main': {'log_file': '/unused/mycli.log', 'log_level': 'NONE'}}
    capture_warning_calls: list[bool] = []
    monkeypatch.setattr(client_commands.logging, 'captureWarnings', lambda value: capture_warning_calls.append(value))
    logger = logging.getLogger('mycli')
    original_handlers = list(logger.handlers)
    try:
        client.initialize_logging()

        added_handlers = [handler for handler in logger.handlers if handler not in original_handlers]
        assert len(added_handlers) == 1
        assert isinstance(added_handlers[0], logging.NullHandler)
        assert logger.level == logging.CRITICAL
        assert capture_warning_calls == [True]
    finally:
        for handler in logger.handlers:
            if handler not in original_handlers:
                logger.removeHandler(handler)


def test_initialize_logging_uses_file_handler(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    log_file = tmp_path / 'mycli.log'
    client = DummyClient()
    client.config = {'main': {'log_file': str(log_file), 'log_level': 'DEBUG'}}
    capture_warning_calls: list[bool] = []
    monkeypatch.setattr(client_commands.logging, 'captureWarnings', lambda value: capture_warning_calls.append(value))
    logger = logging.getLogger('mycli')
    original_handlers = list(logger.handlers)
    try:
        client.initialize_logging()

        added_handlers = [handler for handler in logger.handlers if handler not in original_handlers]
        assert len(added_handlers) == 1
        assert isinstance(added_handlers[0], logging.FileHandler)
        assert logger.level == logging.DEBUG
        assert capture_warning_calls == [True]
    finally:
        for handler in logger.handlers:
            if handler not in original_handlers:
                logger.removeHandler(handler)
                handler.close()


def test_initialize_logging_reports_invalid_log_path() -> None:
    client = DummyClient()
    client.config = {'main': {'log_file': '/does/not/exist/mycli.log', 'log_level': 'INFO'}}

    client.initialize_logging()

    assert client.echo_calls == [(('Error: Unable to open the log file "/does/not/exist/mycli.log".',), {'err': True, 'fg': 'red'})]



from __future__ import annotations

from pathlib import Path
from typing import Any, Iterator, Sequence

import pytest

import mycli.packages.polars_transform as polars_transform
from mycli.packages.polars_transform import (
    PolarsPipeline,
    PolarsTransform,
    PolarsTransformError,
    parse_polars_transform,
    prepare_polars_transform,
    run_polars_transform,
)
from mycli.packages.sqlresult import SQLResult
from mycli.types import ImageProtocol, OutputMode


class FakeDataFrame:
    written_paths: list[str] = []
    written_dataframes: list['FakeDataFrame'] = []

    def __init__(
        self,
        rows: list[tuple[Any, ...]] | list[Any],
        schema: list[str],
        orient: str | None = None,
    ) -> None:
        if orient is None:
            self.rows = [(value,) for value in rows]
        else:
            assert orient == 'row'
            self.rows = rows
        self.columns = schema
        self.parquet_paths: list[str] = []

    def iter_rows(self) -> Iterator[tuple[Any, ...]]:
        return iter(self.rows)

    def __len__(self) -> int:
        return len(self.rows)

    def write_parquet(self, path: str) -> None:
        self.parquet_paths.append(path)
        self.written_paths.append(path)
        self.written_dataframes.append(self)


class FakeSeries:
    def __init__(
        self,
        name: str | Sequence[Any],
        values: list[Any] | None = None,
        *,
        strict: bool = True,
    ) -> None:
        if values is None:
            self.name = ''
            self.values = list(name)
        else:
            assert isinstance(name, str)
            self.name = name
            self.values = values

    def __iter__(self) -> Iterator[Any]:
        return iter(self.values)

    def rename(self, name: str) -> 'FakeSeries':
        return FakeSeries(name, self.values)

    def to_frame(self) -> FakeDataFrame:
        return FakeDataFrame([(value,) for value in self.values], [self.name], 'row')


class FakePolars:
    DataFrame = FakeDataFrame
    Series = FakeSeries


class FailingDataFrame(FakeDataFrame):
    def write_parquet(self, path: str) -> None:
        raise OSError('disk full')


class FailingPolars:
    DataFrame = FailingDataFrame


class ScalarConstructionFailingDataFrame(FakeDataFrame):
    def __init__(
        self,
        rows: list[tuple[Any, ...]] | list[Any],
        schema: list[str],
        orient: str | None = None,
    ) -> None:
        if orient is None:
            raise TypeError('invalid scalar')
        super().__init__(rows, schema, orient)


class ScalarConstructionFailingPolars:
    DataFrame = ScalarConstructionFailingDataFrame
    Series = FakeSeries


class SequenceConstructionFailingSeries(FakeSeries):
    def __init__(
        self,
        name: str | Sequence[Any],
        values: list[Any] | None = None,
        *,
        strict: bool = True,
    ) -> None:
        if values is None:
            raise TypeError('invalid sequence')
        super().__init__(name, values, strict=strict)


class SequenceConstructionFailingPolars:
    DataFrame = FakeDataFrame
    Series = SequenceConstructionFailingSeries


class FakeTopLevelMixin:
    pass


class FakePlot(FakeTopLevelMixin):
    formats: list[str] = []
    scale_factors: list[float] = []
    ppis: list[int] = []
    saved_paths: list[str] = []

    def __init__(self, _data: FakeDataFrame) -> None:
        self.save_calls: list[str] = []

    def save(self, file: Any, **kwargs: Any) -> None:
        self.save_calls.append(kwargs['format'])
        self.formats.append(kwargs['format'])
        if 'scale_factor' in kwargs:
            self.scale_factors.append(float(kwargs['scale_factor']))
        if 'ppi' in kwargs:
            self.ppis.append(kwargs['ppi'])
        if isinstance(file, str):
            self.saved_paths.append(file)
        else:
            file.write(b'png image')


class FailingPlot(FakeTopLevelMixin):
    def save(self, _file: Any, **_kwargs: str) -> None:
        raise OSError('renderer failed')


class FakeTheme:
    enabled: list[str] = []

    @classmethod
    def enable(cls, name: str) -> None:
        cls.enabled.append(name)


class FakeAltair:
    TopLevelMixin = FakeTopLevelMixin
    theme = FakeTheme

    @staticmethod
    def Plot(data: FakeDataFrame) -> FakePlot:
        return FakePlot(data)


class FailingAltair:
    TopLevelMixin = FakeTopLevelMixin
    theme = FakeTheme

    @staticmethod
    def Plot(_data: FakeDataFrame) -> FailingPlot:
        return FailingPlot()


def make_transform(expression: str) -> PolarsTransform:
    return PolarsTransform(
        sql='SELECT id FROM orders',
        expression=expression,
        code=compile(expression, '<test>', 'eval'),
        polars=FakePolars,
        altair=FakeAltair,
    )


def test_parse_polars_transform_splits_sql_and_preserves_expression() -> None:
    assert parse_polars_transform("SELECT * FROM orders .| df.filter(pl.col('state') == 'open')") == PolarsPipeline(
        sql='SELECT * FROM orders',
        expression="df.filter(pl.col('state') == 'open')",
        output_path=None,
        output_mode='tabular',
    )


@pytest.mark.parametrize(
    ('command', 'output_mode'),
    [
        (r'SELECT * FROM orders \x .| df', 'explorer'),
        (r'SELECT * FROM orders .| df \x', 'explorer'),
        (r'SELECT * FROM orders .| df \G', 'expanded'),
    ],
)
def test_parse_polars_transform_returns_output_mode(command: str, output_mode: OutputMode) -> None:
    assert parse_polars_transform(command) == PolarsPipeline(
        sql='SELECT * FROM orders',
        expression='df',
        output_path=None,
        output_mode=output_mode,
    )


@pytest.mark.parametrize(
    ('command', 'expression', 'path'),
    [
        ('SELECT * FROM orders .> orders.parquet', None, 'orders.parquet'),
        ('SELECT * FROM orders .> orders.parquet;', None, 'orders.parquet'),
        (r'SELECT * FROM orders .> orders.parquet \g', None, 'orders.parquet'),
        ("SELECT * FROM orders .> 'order exports.parquet'", None, 'order exports.parquet'),
        ('SELECT * FROM orders .| df.head(10) .> orders.parquet', 'df.head(10)', 'orders.parquet'),
        ('SELECT * FROM orders .| alt.Plot(df) .> orders.png', 'alt.Plot(df)', 'orders.png'),
        (r"SELECT * FROM orders .| alt.Plot(df) .> 'order plot.png' \g", 'alt.Plot(df)', 'order plot.png'),
        ('SELECT * FROM orders .| alt.Plot(df) .> orders.pdf', 'alt.Plot(df)', 'orders.pdf'),
        (r"SELECT * FROM orders .| alt.Plot(df) .> 'order plot.SVG' \g", 'alt.Plot(df)', 'order plot.SVG'),
        (r"SELECT * FROM orders .| alt.Plot(df) .> 'order plot.HTML' \g", 'alt.Plot(df)', 'order plot.HTML'),
    ],
)
def test_parse_polars_transform_parses_file_redirect(
    command: str,
    expression: str | None,
    path: str,
) -> None:
    assert parse_polars_transform(command) == PolarsPipeline(
        sql='SELECT * FROM orders',
        expression=expression,
        output_path=path,
        output_mode='tabular',
    )


@pytest.mark.parametrize(
    'command',
    [
        "SELECT '.|' AS marker",
        'SELECT 1 /* .| df */',
        'SELECT 1.| df',
        'SELECT table.| df',
        'SELECT 1 .|df',
        'SELECT 1 .+ df',
    ],
)
def test_parse_polars_transform_ignores_non_suffix_markers(command: str) -> None:
    assert parse_polars_transform(command) is None


@pytest.mark.parametrize(
    ('command', 'message'),
    [
        (' .| df', 'require a SQL statement'),
        ('SELECT 1 .| ', 'require a Python expression'),
        ('SELECT 1 .|', 'require a Python expression'),
        ('SELECT 1; SELECT 2 .| df', 'exactly one SQL statement'),
        ('SELECT 1 .>', 'require a destination path'),
        ('SELECT 1 .> export.csv', 'end in ".parquet", ".png", ".pdf", ".svg", or ".html"'),
        ('SELECT 1 .> export file.parquet', 'must be quoted'),
        ('SELECT 1 .> export.parquet .| df', 'must follow'),
        ('SELECT 1 .| df .> export.parquet \\x', 'cannot use special display terminators'),
        ('SELECT 1 .| df .> export.parquet \\G', 'cannot use special display terminators'),
        ('SELECT 1 .| df .| df', 'only one ".|"'),
        ('SELECT 1 .> first.parquet .> second.parquet', 'only one ".>"'),
        ('SELECT 1 .> ;', 'require a destination path'),
    ],
)
def test_parse_polars_transform_rejects_invalid_commands(command: str, message: str) -> None:
    with pytest.raises(PolarsTransformError, match=message):
        parse_polars_transform(command)


def test_parse_polars_transform_reports_tokenizer_error(monkeypatch: pytest.MonkeyPatch) -> None:
    def fail_tokenize(command: str) -> list[Any]:
        raise polars_transform.sqlglot.errors.TokenError('bad token')

    monkeypatch.setattr(polars_transform.sqlglot, 'tokenize', fail_tokenize)

    with pytest.raises(PolarsTransformError, match='Unable to parse Polars transform'):
        parse_polars_transform('SELECT 1 .| df')


def test_parse_polars_transform_reports_sql_parser_error(monkeypatch: pytest.MonkeyPatch) -> None:
    def fail_parse(sql: str, *, read: str) -> list[Any]:
        raise polars_transform.sqlglot.errors.ParseError('bad SQL')

    monkeypatch.setattr(polars_transform.sqlglot, 'parse', fail_parse)

    with pytest.raises(PolarsTransformError, match='Unable to parse SQL before Polars transform'):
        parse_polars_transform('SELECT 1 .| df')


def test_parse_output_path_rejects_mismatched_quotes() -> None:
    with pytest.raises(PolarsTransformError, match='matching quotes'):
        polars_transform._parse_output_path("'orders.parquet")


def test_prepare_polars_transform_compiles_expression_and_loads_polars(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr('mycli.packages.polars_transform._load_polars', lambda: FakePolars)

    transform = prepare_polars_transform('SELECT id FROM orders', 'df')

    assert transform.sql == 'SELECT id FROM orders'
    assert transform.expression == 'df'
    assert transform.polars is FakePolars


def test_prepare_polars_transform_rejects_invalid_expression() -> None:
    with pytest.raises(PolarsTransformError, match='Invalid Polars transform expression'):
        prepare_polars_transform('SELECT 1', 'df[')


def test_prepare_polars_transform_reports_missing_dependency(monkeypatch: pytest.MonkeyPatch) -> None:
    def missing_polars() -> Any:
        raise PolarsTransformError('Polars transforms require Polars.')

    monkeypatch.setattr('mycli.packages.polars_transform._load_polars', missing_polars)

    with pytest.raises(PolarsTransformError, match='require Polars'):
        prepare_polars_transform('SELECT 1', 'df')


def test_load_polars_reports_missing_dependency(monkeypatch: pytest.MonkeyPatch) -> None:
    original_import = __import__

    def fail_polars_import(name: str, *args: Any, **kwargs: Any) -> Any:
        if name == 'polars':
            raise ImportError('not installed')
        return original_import(name, *args, **kwargs)

    monkeypatch.setattr('builtins.__import__', fail_polars_import)

    with pytest.raises(PolarsTransformError, match='require Polars'):
        polars_transform._load_polars()


def test_load_polars_returns_imported_dependency(monkeypatch: pytest.MonkeyPatch) -> None:
    original_import = __import__

    def import_fake_polars(name: str, *args: Any, **kwargs: Any) -> Any:
        if name == 'polars':
            return FakePolars
        return original_import(name, *args, **kwargs)

    monkeypatch.setattr('builtins.__import__', import_fake_polars)

    assert polars_transform._load_polars() is FakePolars


def test_load_altair_reports_missing_dependency(monkeypatch: pytest.MonkeyPatch) -> None:
    original_import = __import__

    def fail_altair_import(name: str, *args: Any, **kwargs: Any) -> Any:
        if name == 'altair':
            raise ImportError('not installed')
        return original_import(name, *args, **kwargs)

    monkeypatch.setattr('builtins.__import__', fail_altair_import)

    with pytest.raises(PolarsTransformError, match='require Altair'):
        polars_transform._load_altair()


def test_load_vl_convert_reports_missing_dependency(monkeypatch: pytest.MonkeyPatch) -> None:
    original_import = __import__

    def fail_vl_convert_import(name: str, *args: Any, **kwargs: Any) -> Any:
        if name == 'vl_convert':
            raise ImportError('not installed')
        return original_import(name, *args, **kwargs)

    monkeypatch.setattr('builtins.__import__', fail_vl_convert_import)

    with pytest.raises(PolarsTransformError, match='requires vl-convert-python'):
        polars_transform._load_vl_convert()


def test_run_polars_transform_materializes_and_renders_returned_dataframe() -> None:
    result = run_polars_transform(
        make_transform('df'),
        iter([SQLResult(header=['id'], rows=[(1,), (2,)])]),
    )

    assert result == SQLResult(header=['id'], rows=[(1,), (2,)])


@pytest.mark.parametrize(
    ('expression', 'column_name', 'value'),
    [
        ('42', '42', 42),
        ("'total'", 'total', 'total'),
    ],
)
def test_run_polars_transform_coerces_scalar_to_dataframe(
    expression: str,
    column_name: str,
    value: Any,
) -> None:
    result = run_polars_transform(
        make_transform(expression),
        iter([SQLResult(header=['id'], rows=[(1,)])]),
    )

    assert result == SQLResult(header=[column_name], rows=[(value,)])


def test_run_polars_transform_reports_invalid_scalar() -> None:
    transform = PolarsTransform(
        sql='SELECT id FROM orders',
        expression='42',
        code=compile('42', '<test>', 'eval'),
        polars=ScalarConstructionFailingPolars,
        altair=None,
    )

    with pytest.raises(PolarsTransformError, match='Unable to render scalar as DataFrame: TypeError: invalid scalar'):
        run_polars_transform(
            transform,
            iter([SQLResult(header=['id'], rows=[(1,)])]),
        )


@pytest.mark.parametrize(
    ('expression', 'header', 'rows'),
    [
        ("{'id': [1, 2], 'name': ['one', 'two']}", ['id', 'name'], [(1, 'one'), (2, 'two')]),
        ("{'id': 1, 'name': 'one'}", ['id', 'name'], [(1, 'one')]),
        ('{}', [], []),
    ],
)
def test_run_polars_transform_coerces_dictionary_to_dataframe(
    expression: str,
    header: list[str],
    rows: list[tuple[Any, ...]],
) -> None:
    import polars as pl

    transform = PolarsTransform(
        sql='SELECT id FROM orders',
        expression=expression,
        code=compile(expression, '<test>', 'eval'),
        polars=pl,
        altair=None,
    )

    result = run_polars_transform(
        transform,
        iter([SQLResult(header=['id'], rows=[(1,)])]),
    )

    assert result == SQLResult(header=header, rows=rows)


def test_run_polars_transform_coerces_sequence_to_series() -> None:
    result = run_polars_transform(
        make_transform('[1, None, 3]'),
        iter([SQLResult(header=['id'], rows=[(1,)])]),
    )

    assert result == SQLResult(header=['value'], rows=[(1,), (None,), (3,)])


def test_run_polars_transform_reports_invalid_sequence() -> None:
    transform = PolarsTransform(
        sql='SELECT id FROM orders',
        expression='[1, 2]',
        code=compile('[1, 2]', '<test>', 'eval'),
        polars=SequenceConstructionFailingPolars,
        altair=None,
    )

    with pytest.raises(PolarsTransformError, match='Unable to render Sequence as Series: TypeError: invalid sequence'):
        run_polars_transform(
            transform,
            iter([SQLResult(header=['id'], rows=[(1,)])]),
        )


def test_run_polars_transform_returns_empty_result_for_none() -> None:
    result = run_polars_transform(
        make_transform('None'),
        iter([SQLResult(header=['id'], rows=[(1,)])]),
    )

    assert result == SQLResult()


def test_run_polars_transform_reports_unsupported_return_type() -> None:
    result = run_polars_transform(
        make_transform('object()'),
        iter([SQLResult(header=['id'], rows=[(1,)])]),
    )

    assert result.status_plain == "Nothing could be displayed for return type: <class 'object'>"


def test_run_polars_transform_reports_invalid_dictionary() -> None:
    import polars as pl

    expression = "{'id': [1, 2], 'name': ['one']}"
    transform = PolarsTransform(
        sql='SELECT id FROM orders',
        expression=expression,
        code=compile(expression, '<test>', 'eval'),
        polars=pl,
        altair=None,
    )

    with pytest.raises(PolarsTransformError, match='Unable to render dictionary as DataFrame'):
        run_polars_transform(
            transform,
            iter([SQLResult(header=['id'], rows=[(1,)])]),
        )


def test_run_polars_transform_writes_dictionary_to_parquet(tmp_path: Path) -> None:
    import polars as pl

    path = tmp_path / 'orders.parquet'
    expression = "{'id': [1, 2], 'name': ['one', 'two']}"
    transform = PolarsTransform(
        sql='SELECT id FROM orders',
        expression=expression,
        code=compile(expression, '<test>', 'eval'),
        polars=pl,
        altair=None,
    )

    result = run_polars_transform(
        transform,
        iter([SQLResult(header=['id'], rows=[(1,)])]),
        str(path),
    )

    assert result == SQLResult(status=f'Wrote 2 rows to {path}.')
    assert pl.read_parquet(path).to_dict(as_series=False) == {'id': [1, 2], 'name': ['one', 'two']}


@pytest.mark.parametrize(
    ('expression', 'header', 'rows'),
    [
        ("pl.Series('total', [1, None, 3])", ['total'], [(1,), (None,), (3,)]),
        ("pl.Series('', ['a', 'b'])", ['value'], [('a',), ('b',)]),
    ],
)
def test_run_polars_transform_renders_series(
    expression: str,
    header: list[str],
    rows: list[tuple[Any, ...]],
) -> None:
    result = run_polars_transform(
        make_transform(expression),
        iter([SQLResult(header=['id'], rows=[(1,)])]),
    )

    assert result == SQLResult(header=header, rows=rows)


def test_run_polars_transform_writes_raw_dataframe_to_parquet() -> None:
    transform = make_transform('df')
    FakeDataFrame.written_paths = []
    FakeDataFrame.written_dataframes = []

    result = run_polars_transform(
        transform,
        iter([SQLResult(header=['id'], rows=[(1,), (2,)])]),
        'orders.parquet',
    )

    assert result == SQLResult(status='Wrote 2 rows to orders.parquet.')
    assert FakeDataFrame.written_paths == ['orders.parquet']


@pytest.mark.parametrize(
    ('expression', 'column_name', 'rows'),
    [
        ("pl.Series('total', [1, None, 3])", 'total', [(1,), (None,), (3,)]),
        ("pl.Series('', ['a', 'b'])", 'value', [('a',), ('b',)]),
    ],
)
def test_run_polars_transform_writes_series_to_parquet(
    expression: str,
    column_name: str,
    rows: list[tuple[Any, ...]],
) -> None:
    FakeDataFrame.written_paths = []
    FakeDataFrame.written_dataframes = []

    result = run_polars_transform(
        make_transform(expression),
        iter([SQLResult(header=['id'], rows=[(1,)])]),
        'series.parquet',
    )

    assert result == SQLResult(status=f'Wrote {len(rows)} rows to series.parquet.')
    assert FakeDataFrame.written_paths == ['series.parquet']
    assert FakeDataFrame.written_dataframes[-1].columns == [column_name]
    assert FakeDataFrame.written_dataframes[-1].rows == rows


def test_run_polars_transform_reports_series_parquet_write_error(monkeypatch: pytest.MonkeyPatch) -> None:
    def fail_write(self: FakeDataFrame, path: str) -> None:
        raise OSError('disk full')

    monkeypatch.setattr(FakeDataFrame, 'write_parquet', fail_write)

    with pytest.raises(PolarsTransformError, match='Unable to write Parquet file'):
        run_polars_transform(
            make_transform("pl.Series('total', [1])"),
            iter([SQLResult(header=['id'], rows=[(1,)])]),
            'series.parquet',
        )


def test_run_polars_transform_reports_parquet_write_error() -> None:
    transform = PolarsTransform(
        sql='SELECT id FROM orders',
        expression='df',
        code=compile('df', '<test>', 'eval'),
        polars=FailingPolars,
        altair=FakeAltair,
    )

    with pytest.raises(PolarsTransformError, match='Unable to write Parquet file'):
        run_polars_transform(
            transform,
            iter([SQLResult(header=['id'], rows=[(1,)])]),
            'orders.parquet',
        )


def test_run_polars_transform_reports_disabled_altair_plot_output() -> None:
    result = run_polars_transform(
        make_transform('alt.Plot(df)'),
        iter([SQLResult(header=['id'], rows=[(1,)])]),
    )

    assert result.status_plain == 'image_protocol is unset in ~/.myclirc. Inline plotting is disabled.'


@pytest.mark.parametrize('image_protocol', ['iterm2', 'kitty'])
def test_run_polars_transform_renders_altair_plot_for_image_protocol(
    monkeypatch: pytest.MonkeyPatch,
    image_protocol: ImageProtocol,
) -> None:
    load_calls: list[None] = []
    FakePlot.scale_factors = []
    FakePlot.ppis = []
    FakeTheme.enabled = []
    monkeypatch.setattr(polars_transform, '_load_vl_convert', lambda: load_calls.append(None))

    result = run_polars_transform(
        make_transform('alt.Plot(df)'),
        iter([SQLResult(header=['id'], rows=[(1,)])]),
        image_protocol=image_protocol,
        plot_scale_factor=1.5,
        plot_ppi=144,
        plot_theme='dark',
    )

    assert load_calls == [None]
    assert FakePlot.scale_factors == [1.5]
    assert FakePlot.ppis == [144]
    assert FakeTheme.enabled == ['dark']
    assert result == SQLResult(image=b'png image', image_protocol=image_protocol)


@pytest.mark.parametrize(
    ('path', 'plot_format', 'expected_scale_factors', 'expected_ppis', 'expected_load_calls', 'output_kind'),
    [
        ('plot.png', 'png', [1.5], [144], [None], 'image'),
        ('plot.pdf', 'pdf', [1.5], [], [None], 'image'),
        ('plot.SVG', 'svg', [1.5], [], [None], 'image'),
        ('plot.html', 'html', [], [], [], 'document'),
    ],
)
def test_run_polars_transform_writes_altair_plot_to_file(
    monkeypatch: pytest.MonkeyPatch,
    path: str,
    plot_format: str,
    expected_scale_factors: list[float],
    expected_ppis: list[int],
    expected_load_calls: list[None],
    output_kind: str,
) -> None:
    load_calls: list[None] = []
    FakePlot.formats = []
    FakePlot.scale_factors = []
    FakePlot.ppis = []
    FakePlot.saved_paths = []
    FakeTheme.enabled = []
    monkeypatch.setattr(polars_transform, '_load_vl_convert', lambda: load_calls.append(None))

    result = run_polars_transform(
        make_transform('alt.Plot(df)'),
        iter([SQLResult(header=['id'], rows=[(1,)])]),
        path,
        image_protocol='none',
        plot_scale_factor=1.5,
        plot_ppi=144,
        plot_theme='dark',
    )

    assert result == SQLResult(status=f'Wrote {plot_format.upper()} {output_kind} to {path}.')
    assert load_calls == expected_load_calls
    assert FakePlot.formats == [plot_format]
    assert FakePlot.saved_paths == [path]
    assert FakePlot.scale_factors == expected_scale_factors
    assert FakePlot.ppis == expected_ppis
    assert FakeTheme.enabled == ['dark']


@pytest.mark.parametrize('plot_format', ['png', 'pdf', 'svg', 'html'])
def test_run_polars_transform_reports_altair_file_write_error(
    monkeypatch: pytest.MonkeyPatch,
    plot_format: str,
) -> None:
    monkeypatch.setattr(polars_transform, '_load_vl_convert', lambda: None)
    transform = PolarsTransform(
        sql='SELECT id FROM orders',
        expression='alt.Plot(df)',
        code=compile('alt.Plot(df)', '<test>', 'eval'),
        polars=FakePolars,
        altair=FailingAltair,
    )

    path = f'plot.{plot_format}'
    with pytest.raises(
        PolarsTransformError,
        match=f'Unable to write {plot_format.upper()} file "{path}": OSError: renderer failed',
    ):
        run_polars_transform(
            transform,
            iter([SQLResult(header=['id'], rows=[(1,)])]),
            path,
        )


def test_run_polars_transform_uses_integer_default_plot_ppi(monkeypatch: pytest.MonkeyPatch) -> None:
    FakePlot.ppis = []
    monkeypatch.setattr(polars_transform, '_load_vl_convert', lambda: None)

    run_polars_transform(
        make_transform('alt.Plot(df)'),
        iter([SQLResult(header=['id'], rows=[(1,)])]),
        image_protocol='kitty',
    )

    assert FakePlot.ppis == [200]


def test_run_polars_transform_reports_invalid_altair_theme(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(polars_transform, '_load_vl_convert', lambda: None)

    def fail_enable(_name: str) -> None:
        raise ValueError('unknown theme')

    monkeypatch.setattr(FakeTheme, 'enable', fail_enable)

    with pytest.raises(PolarsTransformError, match='Unable to enable Altair plot theme "not-a-theme": ValueError: unknown theme'):
        run_polars_transform(
            make_transform('alt.Plot(df)'),
            iter([SQLResult(header=['id'], rows=[(1,)])]),
            image_protocol='kitty',
            plot_theme='not-a-theme',
        )


def test_run_polars_transform_reports_altair_plot_render_error(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(polars_transform, '_load_vl_convert', lambda: None)
    transform = PolarsTransform(
        sql='SELECT id FROM orders',
        expression='alt.Plot(df)',
        code=compile('alt.Plot(df)', '<test>', 'eval'),
        polars=FakePolars,
        altair=FailingAltair,
    )

    with pytest.raises(PolarsTransformError, match='Unable to render Altair plot: OSError: renderer failed'):
        run_polars_transform(
            transform,
            iter([SQLResult(header=['id'], rows=[(1,)])]),
            image_protocol='iterm2',
        )


def test_run_polars_transform_reports_missing_altair_renderer(monkeypatch: pytest.MonkeyPatch) -> None:
    def fail_load() -> None:
        raise PolarsTransformError('Altair plot rendering requires vl-convert-python. Install mycli[dataframe].')

    monkeypatch.setattr(polars_transform, '_load_vl_convert', fail_load)

    with pytest.raises(PolarsTransformError, match='requires vl-convert-python'):
        run_polars_transform(
            make_transform('alt.Plot(df)'),
            iter([SQLResult(header=['id'], rows=[(1,)])]),
            image_protocol='iterm2',
        )


def test_run_polars_transform_rejects_altair_plot_parquet_output() -> None:
    with pytest.raises(PolarsTransformError, match='Altair plots can only be written'):
        run_polars_transform(
            make_transform('alt.Plot(df)'),
            iter([SQLResult(header=['id'], rows=[(1,)])]),
            'plot.parquet',
            image_protocol='iterm2',
        )


@pytest.mark.parametrize(
    ('expression', 'path', 'message'),
    [
        ('df', 'data.pdf', 'DataFrame results can only be written'),
        ("pl.Series('id', [1])", 'data.svg', 'Series results can only be written'),
        ('df', 'data.html', 'DataFrame results can only be written'),
    ],
)
def test_run_polars_transform_rejects_dataframe_and_series_plot_output(
    expression: str,
    path: str,
    message: str,
) -> None:
    with pytest.raises(PolarsTransformError, match=message):
        run_polars_transform(
            make_transform(expression),
            iter([SQLResult(header=['id'], rows=[(1,)])]),
            path,
        )


def test_prepare_polars_transform_supports_direct_parquet_output(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr('mycli.packages.polars_transform._load_polars', lambda: FakePolars)

    transform = prepare_polars_transform('SELECT id FROM orders', None)

    assert transform.expression is None
    assert transform.altair is None


@pytest.mark.parametrize(
    'results',
    [
        iter([]),
        iter([SQLResult(status='OK')]),
        iter([SQLResult(header=['id'], rows=[(1,)]), SQLResult(status='OK')]),
    ],
)
def test_run_polars_transform_rejects_non_single_tabular_result(results: Iterator[SQLResult]) -> None:
    with pytest.raises(PolarsTransformError):
        run_polars_transform(make_transform('df'), results)


def test_run_polars_transform_reports_expression_error() -> None:
    with pytest.raises(PolarsTransformError, match='ZeroDivisionError'):
        run_polars_transform(make_transform('1 / 0'), iter([SQLResult(header=['id'], rows=[(1,)])]))



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from typing import Literal

import click

PasswordSource = Literal[
    'prompt',
    'literal',
    'file',
    'environment',
    'dsn',
    'boundary',
    'vault',
    'login_path',
    'keyring',
    'fallback',
]
PasswordValue = str | int
PasswordLoader = Callable[[], PasswordValue | None]

KNOWN_PASSWORD_SOURCES: list[PasswordSource] = [
    'prompt',
    'literal',
    'file',
    'environment',
    'dsn',
    'boundary',
    'vault',
    'login_path',
    'keyring',
    'fallback',
]


@dataclass(frozen=True, slots=True)
class SelectedPassword:
    source: PasswordSource
    value: PasswordValue


@dataclass(frozen=True, slots=True)
class PasswordCandidate:
    source: PasswordSource
    value: PasswordValue | None = None
    loader: PasswordLoader | None = None

    def resolve(self) -> SelectedPassword | None:
        value = self.loader() if self.loader is not None else self.value
        if value is None:
            return None
        return SelectedPassword(source=self.source, value=value)


class PasswordCandidates:
    def __init__(self) -> None:
        self._candidates: dict[PasswordSource, PasswordCandidate] = {}

    def add_value(self, source: PasswordSource, value: PasswordValue | None) -> None:
        self._candidates[source] = PasswordCandidate(source=source, value=value)

    def add_loader(self, source: PasswordSource, loader: PasswordLoader) -> None:
        self._candidates[source] = PasswordCandidate(source=source, loader=loader)

    def resolve(self, password_source_precedence) -> SelectedPassword | None:
        for source in password_source_precedence:
            if source not in KNOWN_PASSWORD_SOURCES:
                click.secho(f'Skipping unknown password source: {source}.', err=True, fg='red')
                continue
            candidate = self._candidates.get(source)
            if candidate is None:
                continue
            if selected := candidate.resolve():
                return selected
        return None



from __future__ import annotations

from dataclasses import dataclass
from io import TextIOWrapper
import os
import sys
from textwrap import dedent
from typing import Callable

import click
import clickdc

import mycli as mycli_package
from mycli.cli_runner import run_from_cli_args
from mycli.client import MyCli
from mycli.constants import (
    DEFAULT_PROMPT,
    EMPTY_PASSWORD_FLAG_SENTINEL,
)
from mycli.packages.cli_utils import filtered_sys_argv


class IntOrStringClickParamType(click.ParamType):
    name = 'text'  # display as TEXT in helpdoc

    def convert(self, value, param, ctx):
        if isinstance(value, int):
            return value
        elif isinstance(value, str):
            return value
        elif value is None:
            return value
        else:
            self.fail('Not a valid password string', param, ctx)


INT_OR_STRING_CLICK_TYPE = IntOrStringClickParamType()


@dataclass(slots=True)
class CliArgs:
    database: str | None = clickdc.argument(
        type=str,
        default=None,
        nargs=1,
    )
    host: str | None = clickdc.option(
        '-h',
        '--hostname',
        'host',
        type=str,
        envvar='MYSQL_HOST',
        help='Host address of the database.',
    )
    port: int | None = clickdc.option(
        '-P',
        type=int,
        envvar='MYSQL_TCP_PORT',
        help='Port number to use for connection. Honors $MYSQL_TCP_PORT.',
    )
    user: str | None = clickdc.option(
        '-u',
        '--user',
        '--username',
        'user',
        type=str,
        envvar='MYSQL_USER',
        help='User name to connect to the database.',
    )
    socket: str | None = clickdc.option(
        '-S',
        type=str,
        envvar='MYSQL_UNIX_SOCKET',
        help='The socket file to use for connection.',
    )
    password: int | str | None = clickdc.option(
        '-p',
        '--pass',
        '--password',
        'password',
        type=INT_OR_STRING_CLICK_TYPE,
        help=dedent(
            """Password to connect to the database.
            Use with a value to set the password at the CLI, or alone in the last position to request a prompt.
            """
        ),
    )
    password_file: str | None = clickdc.option(
        type=click.Path(),
        help='File or FIFO path containing the password to connect to the db if not specified otherwise.',
    )
    vault_address: str | None = clickdc.option(
        type=str,
        help='EXPERIMENTAL "vault kv get" integration: value for $VAULT_ADDR if unset in the environment or ~/.myclirc.',
    )
    vault_mount: str | None = clickdc.option(
        type=str,
        help='EXPERIMENTAL "vault kv get" integration: value for -mount if unset in ~/.myclirc.',
    )
    vault_secret: str | None = clickdc.option(
        type=str,
        help='EXPERIMENTAL "vault kv get" integration: secret name.',
    )
    vault_password_field: str | None = clickdc.option(
        type=str,
        help='EXPERIMENTAL "vault kv get" integration: field containing the password.',
    )
    vault_username_field: str | None = clickdc.option(
        type=str,
        help='EXPERIMENTAL "vault kv get" integration: field containing the username.',
    )
    ssl_mode: str = clickdc.option(
        type=click.Choice(['auto', 'on', 'off']),
        help='Set desired SSL behavior. auto=preferred if TCP/IP, on=required, off=off.',
    )
    ssl_ca: str | None = clickdc.option(
        type=click.Path(exists=True),
        help='CA file in PEM format.',
    )
    ssl_capath: str | None = clickdc.option(
        type=click.Path(exists=True, file_okay=False, dir_okay=True),
        help='CA directory.',
    )
    ssl_cert: str | None = clickdc.option(
        type=click.Path(exists=True),
        help='X509 cert in PEM format.',
    )
    ssl_key: str | None = clickdc.option(
        type=click.Path(exists=True),
        help='X509 key in PEM format.',
    )
    ssl_cipher: str | None = clickdc.option(
        type=str,
        help='SSL cipher to use.',
    )
    tls_version: str | None = clickdc.option(
        type=click.Choice(['TLSv1', 'TLSv1.1', 'TLSv1.2', 'TLSv1.3'], case_sensitive=False),
        help='TLS protocol version for secure connection.',
    )
    ssl_verify_server_cert: bool = clickdc.option(
        is_flag=True,
        help=("""Verify server's "Common Name" in its cert against hostname used when connecting. This option is disabled by default."""),
    )
    verbose: int = clickdc.option(
        '-v',
        count=True,
        help='More verbose output and feedback.  Can be given multiple times.',
    )
    quiet: bool = clickdc.option(
        '-q',
        is_flag=True,
        help='Less verbose output and feedback.',
    )
    dbname: str | None = clickdc.option(
        '-D',
        '--database',
        'dbname',
        type=str,
        clickdc=None,
        help='Database or DSN to use for the connection.',
    )
    dsn: str = clickdc.option(
        '-d',
        type=str,
        default='',
        envvar='MYSQL_DSN',
        help='DSN alias configured in the ~/.myclirc file, or a full DSN.',
    )
    list_dsn: bool = clickdc.option(
        is_flag=True,
        help='Show list of DSN aliases configured in the [alias_dsn] section of ~/.myclirc.',
    )
    prompt: str | None = clickdc.option(
        '-R',
        type=str,
        help=f'Prompt format (Default: "{DEFAULT_PROMPT}").',
    )
    toolbar: str | None = clickdc.option(
        type=str,
        help='Toolbar format.',
    )
    logfile: TextIOWrapper | None = clickdc.option(
        '-l',
        type=click.File(mode='a', encoding='utf-8'),
        help='Log every query and its results to a file.',
    )
    checkpoint: str | None = clickdc.option(
        type=str,
        help='In batch or --execute mode, log successful queries to a file, and skip them with --resume.',
    )
    resume: bool = clickdc.option(
        '--resume',
        is_flag=True,
        help='In batch mode, resume after replaying statements in the --checkpoint file.',
    )
    myclirc: str = clickdc.option(
        type=click.Path(),
        help='Location of myclirc file.',
    )
    auto_vertical_output: bool = clickdc.option(
        is_flag=True,
        help='Automatically switch to vertical output mode if the result is wider than the terminal width.',
    )
    show_warnings: bool | None = clickdc.option(
        '--show-warnings/--no-show-warnings',
        is_flag=True,
        default=None,
        clickdc=None,
        help='Automatically show warnings after executing a SQL statement.',
    )
    table: bool = clickdc.option(
        '-t',
        is_flag=True,
        help='Shorthand for --format=table.',
    )
    csv: bool = clickdc.option(
        is_flag=True,
        help='Shorthand for --format=csv.',
    )
    warn: bool | None = clickdc.option(
        '--warn/--no-warn',
        default=None,
        clickdc=None,
        help='Warn before running a destructive query.',
    )
    warn_batch: bool = clickdc.option(
        is_flag=True,
        help='Warn before running a destructive query when executing a script.',
    )
    local_infile: bool | None = clickdc.option(
        type=bool,
        is_flag=False,
        default=None,
        help='Enable/disable LOAD DATA LOCAL INFILE.',
    )
    login_path: str | None = clickdc.option(
        '-g',
        type=str,
        help='Read this path from the login file.',
    )
    execute: str | None = clickdc.option(
        '-e',
        type=str,
        help='Execute command and quit.',
    )
    init_command: str | None = clickdc.option(
        type=str,
        help='SQL statement to execute after connecting.',
    )
    unbuffered: bool | None = clickdc.option(
        is_flag=True,
        help='Instead of copying every row of data into a buffer, fetch rows as needed, to save memory.',
    )
    character_set: str | None = clickdc.option(
        '--charset',
        '--character-set',
        'character_set',
        type=str,
        help='Character set for MySQL session.',
    )
    batch: str | None = clickdc.option(
        type=str,
        help='SQL script to execute in batch mode.',
    )
    # deprecated 2026-06-20
    noninteractive: bool = clickdc.option(
        is_flag=True,
        hidden=True,
        deprecated='See --warn-batch.',
    )
    format: str | None = clickdc.option(
        type=click.Choice(['default', 'csv', 'tsv', 'table']),
        help='Format for batch or --execute output.',
    )
    throttle: float = clickdc.option(
        type=float,
        default=0.0,
        help='Pause in seconds between queries in batch mode.',
    )
    progress: bool = clickdc.option(
        is_flag=True,
        help='Show progress on the standard error with --batch.',
    )
    use_keyring: str | None = clickdc.option(
        type=click.Choice(['auto', 'true', 'false', 'reset']),
        default=None,
        help='Store and retrieve passwords from the system keyring. auto means true, unless within an SSH connection.',
    )
    keepalive_ticks: int | None = clickdc.option(
        type=int,
        help='Send regular keepalive pings to the connection, roughly every <int> seconds.',
    )
    ssh_jump: str | None = clickdc.option(
        type=str,
        help='Open an SSH tunnel via [user@]host[:port] and connect to MySQL through it.',
    )
    ssh_options: str | None = clickdc.option(
        type=str,
        help='Extra CLI arguments for SSH with --ssh-jump, placed after options from myclirc.',
    )
    checkup: bool = clickdc.option(
        is_flag=True,
        help='Run a checkup on your configuration.',
    )
    # hidden options which have no effect as of mycli 2.0.0, 2026-07.
    # todo: remove the hidden options, since they are still advertised
    # in spelling corrections.
    ssl: bool | None = clickdc.option(
        '--ssl/--no-ssl',
        clickdc=None,
        hidden=True,
        deprecated='No effect. See --ssl-mode.',
    )
    ssh_user: str | None = clickdc.option(
        type=str,
        hidden=True,
        deprecated='No effect. See --ssh-jump.',
    )
    ssh_host: str | None = clickdc.option(
        type=str,
        hidden=True,
        deprecated='No effect. See --ssh-jump.',
    )
    ssh_port: int = clickdc.option(
        type=int,
        hidden=True,
        deprecated='No effect. See --ssh-jump.',
    )
    ssh_password: str | None = clickdc.option(
        type=str,
        hidden=True,
        deprecated='No effect. See --ssh-jump.',
    )
    ssh_key_filename: str | None = clickdc.option(
        type=str,
        hidden=True,
        deprecated='No effect. See --ssh-jump.',
    )
    ssh_config_path: str = clickdc.option(
        type=str,
        hidden=True,
        deprecated='No effect. See --ssh-jump.',
    )
    ssh_config_host: str | None = clickdc.option(
        type=str,
        hidden=True,
        deprecated='No effect. See --ssh-jump.',
    )
    list_ssh_config: bool = clickdc.option(
        is_flag=True,
        hidden=True,
        deprecated='No effect. See --ssh-jump.',
    )
    ssh_warning_off: bool = clickdc.option(
        is_flag=True,
        hidden=True,
        deprecated='No effect. See --ssh-jump.',
    )


def get_password_from_file(password_file: str | None) -> str | None:
    if not password_file:
        return None
    try:
        with open(password_file) as fp:
            return fp.readline().removesuffix('\n')
    except FileNotFoundError:
        click.secho(f"Password file '{password_file}' not found", err=True, fg='red')
        sys.exit(1)
    except PermissionError:
        click.secho(f"Permission denied reading password file '{password_file}'", err=True, fg='red')
        sys.exit(1)
    except IsADirectoryError:
        click.secho(f"Path '{password_file}' is a directory, not a file", err=True, fg='red')
        sys.exit(1)
    except Exception as e:
        click.secho(f"Error reading password file '{password_file}': {str(e)}", err=True, fg='red')
        sys.exit(1)


def preprocess_cli_args(
    cli_args: CliArgs,
    is_valid_connection_scheme: Callable[[str], tuple[bool, str | None]],
) -> int:
    if cli_args.database is None and isinstance(cli_args.password, str) and '://' in cli_args.password:
        is_valid_scheme, scheme = is_valid_connection_scheme(cli_args.password)
        if not is_valid_scheme:
            click.secho(f'Error: Unknown connection scheme provided for DSN URI ({scheme}://)', err=True, fg='red')
            sys.exit(1)
        cli_args.database = cli_args.password
        cli_args.password = EMPTY_PASSWORD_FLAG_SENTINEL

    if cli_args.resume and not cli_args.checkpoint:
        click.secho('Error: --resume requires a --checkpoint file.', err=True, fg='red')
        sys.exit(1)

    if cli_args.resume and not cli_args.batch:
        click.secho('Error: --resume requires a --batch file.', err=True, fg='red')
        sys.exit(1)

    if (
        cli_args.checkpoint
        and os.path.exists(cli_args.checkpoint)
        and cli_args.batch
        and cli_args.batch != '-'
        and os.path.exists(cli_args.batch)
    ):
        if os.path.samefile(cli_args.batch, cli_args.checkpoint):
            click.secho('Error: --batch and --checkpoint must be different files.', err=True, fg='red')
            sys.exit(1)

    if (
        cli_args.logfile
        and os.path.exists(cli_args.logfile.name)
        and cli_args.batch
        and cli_args.batch != '-'
        and os.path.exists(cli_args.batch)
    ):
        if os.path.samefile(cli_args.batch, cli_args.logfile.name):
            click.secho('Error: --batch and --logfile must be different files.', err=True, fg='red')
            sys.exit(1)

    if cli_args.verbose and cli_args.quiet:
        click.secho('Error: --verbose and --quiet are incompatible.', err=True, fg='red')
        sys.exit(1)
    elif cli_args.verbose:
        return int(cli_args.verbose)
    elif cli_args.quiet:
        return -1
    return 0


@click.command()
@clickdc.adddc('cli_args', CliArgs)
@click.version_option(mycli_package.__version__, '--version', '-V', help="Output mycli's version.")
def click_entrypoint(
    cli_args: CliArgs,
) -> None:
    """A MySQL terminal client with auto-completion and syntax highlighting.

    \b
    Examples:
      - mycli my_database
      - mycli -u my_user -h my_host.com my_database
      - mycli mysql://my_user@my_host.com:3306/my_database

    """

    run_from_cli_args(cli_args, client_factory=MyCli)


def main() -> int | None:
    try:
        result = click_entrypoint.main(
            filtered_sys_argv(),  # type: ignore[arg-type]
            standalone_mode=False,  # disable builtin exception handling
            prog_name='mycli',
        )
    except click.Abort:
        print('Aborted!', file=sys.stderr)
        sys.exit(1)
    except BrokenPipeError:
        sys.exit(1)
    except click.ClickException as e:
        e.show()
        if hasattr(e, 'exit_code'):
            sys.exit(e.exit_code)
        else:
            sys.exit(2)
    if result is None:
        return 0
    elif isinstance(result, int):
        return result
    else:
        return 1


if __name__ == "__main__":
    sys.exit(main())



"""Platform and Python version compatibility support."""

from importlib import import_module
import sys

WIN: bool = sys.platform in ("win32", "cygwin")


def _is_win32() -> bool:
    return sys.platform == 'win32'


def is_windows_console(output: object | None) -> bool:
    """Return whether output uses a native Windows console backend."""
    if not _is_win32() or output is None:
        return False

    output_types = tuple(
        getattr(import_module(module_name), class_name)
        for module_name, class_name in (
            ('prompt_toolkit.output.conemu', 'ConEmuOutput'),
            ('prompt_toolkit.output.win32', 'Win32Output'),
            ('prompt_toolkit.output.windows10', 'Windows10_Output'),
        )
    )

    return isinstance(output, output_types)



from __future__ import annotations

import builtins
from dataclasses import dataclass
import datetime
import decimal
import fractions
from io import BytesIO
from types import CodeType
from typing import Any, Iterable, Sequence

import sqlglot

from mycli.packages.special.delimitercommand import DelimiterCommand
from mycli.packages.sqlresult import SQLResult
from mycli.types import ImageProtocol, OutputMode

delimiter_command = DelimiterCommand()
PLOT_FORMATS = ('png', 'pdf', 'svg', 'html')


class PolarsTransformError(RuntimeError):
    pass


@dataclass(frozen=True, slots=True)
class PolarsPipeline:
    sql: str
    expression: str | None
    output_path: str | None
    output_mode: OutputMode


@dataclass(frozen=True, slots=True)
class PolarsTransform:
    sql: str
    expression: str | None
    code: CodeType
    polars: Any
    altair: Any | None


def _is_coerceable_scalarish(value: Any) -> bool:
    # pretend str/bytes are scalars
    return isinstance(
        value,
        (
            bool,
            int,
            float,
            complex,
            bytes,
            str,
            datetime.datetime,
            datetime.timedelta,
            decimal.Decimal,
            fractions.Fraction,
        ),
    )


def parse_polars_transform(command: str) -> PolarsPipeline | None:
    """Parse a SQL statement with optional Polars transform and file output."""
    try:
        tokens = sqlglot.tokenize(command)
    except sqlglot.errors.TokenError as exc:
        raise PolarsTransformError(f'Unable to parse Polars transform: {exc}') from exc

    pipe_index: int | None = None
    parquet_index: int | None = None
    for index, token in enumerate(tokens[:-1]):
        following = tokens[index + 1]
        if token.token_type != sqlglot.TokenType.DOT:
            continue
        if token.start == 0 or not command[token.start - 1].isspace():
            continue
        if following.token_type not in (sqlglot.TokenType.PIPE, sqlglot.TokenType.GT):
            continue
        if following.end + 1 >= len(command):
            if following.token_type == sqlglot.TokenType.PIPE:
                raise PolarsTransformError('Polars transforms require a Python expression.')
            raise PolarsTransformError('File saves require a destination path.')
        if not command[following.end + 1].isspace():
            continue
        if following.token_type == sqlglot.TokenType.PIPE:
            if pipe_index is not None:
                raise PolarsTransformError('Polars transforms support only one ".|" operator.')
            pipe_index = index
        else:
            if parquet_index is not None:
                raise PolarsTransformError('File saves support only one ".>" operator.')
            parquet_index = index

    if pipe_index is None and parquet_index is None:
        return None
    if parquet_index is not None and pipe_index is not None and parquet_index < pipe_index:
        raise PolarsTransformError('The ".>" operator must follow the ".|" operator.')

    first_index = pipe_index if pipe_index is not None else parquet_index
    assert first_index is not None
    sql = command[: tokens[first_index].start].strip()
    expression: str | None = None
    output_path: str | None = None
    if pipe_index is not None:
        pipe_operator = tokens[pipe_index + 1]
        expression_end = tokens[parquet_index].start if parquet_index is not None else len(command)
        expression = command[pipe_operator.end + 1 : expression_end].strip()
    if parquet_index is not None:
        parquet_operator = tokens[parquet_index + 1]
        output_path = command[parquet_operator.end + 1 :].strip()
        output_path = output_path.removesuffix(delimiter_command.current).rstrip()
        output_path = output_path.removesuffix(r'\g').rstrip()

    has_display_terminator = any(value is not None and value.endswith((r'\x', r'\G')) for value in (sql, expression, output_path))
    if output_path is not None and has_display_terminator:
        raise PolarsTransformError('File saves cannot use special display terminators.')
    if sql.endswith(r'\x') or expression is not None and expression.endswith(r'\x'):
        output_mode: OutputMode = 'explorer'
    elif sql.endswith(r'\G') or expression is not None and expression.endswith(r'\G'):
        output_mode = 'expanded'
    else:
        output_mode = 'tabular'

    for delimiter in (delimiter_command.current, r'\G', r'\g', r'\x'):
        sql = sql.removesuffix(delimiter).rstrip()
        if expression is not None:
            expression = expression.removesuffix(delimiter).rstrip()

    if not sql:
        raise PolarsTransformError('Polars transforms require a SQL statement.')
    if expression is not None and not expression:
        raise PolarsTransformError('Polars transforms require a Python expression.')
    if output_path is not None:
        if not output_path:
            raise PolarsTransformError('File saves require a destination path.')
        output_path = _parse_output_path(output_path)
    _validate_sql(sql)
    return PolarsPipeline(sql=sql, expression=expression, output_path=output_path, output_mode=output_mode)


def _parse_output_path(path: str) -> str:
    if path[0] in ('\'', '"'):
        if len(path) < 2 or path[-1] != path[0]:
            raise PolarsTransformError('File save paths must use matching quotes.')
        path = path[1:-1]
    elif any(character.isspace() for character in path):
        raise PolarsTransformError('File save paths containing spaces must be quoted.')
    if not path.lower().endswith(('.parquet', '.png', '.pdf', '.svg', '.html')):
        raise PolarsTransformError('File save paths must end in ".parquet", ".png", ".pdf", ".svg", or ".html".')
    return path


def _plot_format_for_path(path: str) -> str | None:
    path = path.lower()
    return next((plot_format for plot_format in PLOT_FORMATS if path.endswith(f'.{plot_format}')), None)


def _validate_sql(sql: str) -> None:
    try:
        statements = sqlglot.parse(sql, read='mysql')
    except sqlglot.errors.ParseError as exc:
        raise PolarsTransformError(f'Unable to parse SQL before Polars transform: {exc}') from exc
    if len(statements) != 1:
        raise PolarsTransformError('Polars transforms support exactly one SQL statement.')


def prepare_polars_transform(sql: str, expression: str | None) -> PolarsTransform:
    """Compile a transform expression and load its optional dependency."""
    try:
        code = (
            compile(expression, '<mycli Polars transform>', 'eval')
            if expression is not None
            else compile('df', '<mycli Polars transform>', 'eval')
        )
    except SyntaxError as exc:
        raise PolarsTransformError(f'Invalid Polars transform expression: "{expression}": {exc.msg}') from exc
    return PolarsTransform(
        sql=sql,
        expression=expression,
        code=code,
        polars=_load_polars(),
        altair=_load_altair() if expression is not None else None,
    )


def _load_polars() -> Any:
    try:
        import polars as pl
    except ImportError as exc:
        raise PolarsTransformError("Polars transforms require Polars to be installed.") from exc
    return pl


def _load_altair() -> Any:
    try:
        import altair as alt
    except ImportError as exc:
        raise PolarsTransformError("Polars transforms require Altair to be installed.") from exc
    return alt


def _load_vl_convert() -> None:
    try:
        import vl_convert  # noqa: F401
    except ImportError as exc:
        raise PolarsTransformError('Altair plot rendering requires vl-convert-python. Install mycli[dataframe].') from exc


def run_polars_transform(
    transform: PolarsTransform,
    results: Iterable[SQLResult],
    output_path: str | None = None,
    *,
    image_protocol: ImageProtocol = 'none',
    plot_scale_factor: float = 1.0,
    plot_ppi: int = 200,
    plot_theme: str = 'carbong90',
) -> SQLResult:
    iterator = iter(results)
    try:
        result = next(iterator)
    except StopIteration as exc:
        raise PolarsTransformError('Polars transforms require a tabular SQL result.') from exc
    try:
        next(iterator)
    except StopIteration:
        pass
    else:
        raise PolarsTransformError('Polars transforms do not support multiple result sets.')

    if not isinstance(result.header, list) or result.rows is None:
        raise PolarsTransformError('Polars transforms require a tabular SQL result.')

    dataframe = transform.polars.DataFrame(list(result.rows), schema=result.header, orient='row')
    try:
        value = eval(
            transform.code,
            {
                '__builtins__': builtins,
                'df': dataframe,
                'pl': transform.polars,
                'alt': transform.altair,
            },
        )
    except Exception as exc:
        raise PolarsTransformError(f'Polars expression failed: {type(exc).__name__}: {exc}') from exc

    if _is_coerceable_scalarish(value):
        try:
            value = transform.polars.DataFrame([value], schema=[str(value)])
        except Exception as exc:
            raise PolarsTransformError(f'Unable to render scalar as DataFrame: {type(exc).__name__}: {exc}') from exc
    elif isinstance(value, dict):
        try:
            value = transform.polars.DataFrame(value)
        except Exception as exc:
            raise PolarsTransformError(f'Unable to render dictionary as DataFrame: {type(exc).__name__}: {exc}') from exc
    elif isinstance(value, Sequence):
        try:
            value = transform.polars.Series(value, strict=False)
        except Exception as exc:
            raise PolarsTransformError(f'Unable to render Sequence as Series: {type(exc).__name__}: {exc}') from exc

    if isinstance(value, transform.polars.DataFrame):
        if output_path is not None:
            if not output_path.lower().endswith('.parquet'):
                raise PolarsTransformError('Polars DataFrame results can only be written to ".parquet" files.')
            try:
                value.write_parquet(output_path)
            except Exception as exc:
                raise PolarsTransformError(f'Unable to write Parquet file "{output_path}": {type(exc).__name__}: {exc}') from exc
            return SQLResult(status=f'Wrote {len(value)} rows to {output_path}.')
        return SQLResult(header=list(value.columns), rows=list(value.iter_rows()))
    elif isinstance(value, transform.polars.Series):
        column_name = value.name or 'value'
        if output_path is not None:
            if not output_path.lower().endswith('.parquet'):
                raise PolarsTransformError('Polars Series results can only be written to ".parquet" files.')
            try:
                series_dataframe = value.rename(column_name).to_frame()
                series_dataframe.write_parquet(output_path)
            except Exception as exc:
                raise PolarsTransformError(f'Unable to write Parquet file "{output_path}": {type(exc).__name__}: {exc}') from exc
            return SQLResult(status=f'Wrote {len(series_dataframe)} rows to {output_path}.')
        return SQLResult(header=[column_name], rows=[(item,) for item in value])
    elif transform.altair is not None and isinstance(value, transform.altair.TopLevelMixin):
        plot_format = _plot_format_for_path(output_path) if output_path is not None else None
        if output_path is not None and plot_format is None:
            raise PolarsTransformError('Altair plots can only be written to ".png", ".pdf", ".svg", or ".html" files.')
        if output_path is None and image_protocol == 'none':
            return SQLResult(status='image_protocol is unset in ~/.myclirc. Inline plotting is disabled.')
        if plot_format != 'html':
            _load_vl_convert()
        try:
            transform.altair.theme.enable(plot_theme)
        except Exception as exc:
            raise PolarsTransformError(f'Unable to enable Altair plot theme "{plot_theme}": {type(exc).__name__}: {exc}') from exc
        if output_path is not None:
            assert plot_format is not None
            save_kwargs: dict[str, Any] = {
                'format': plot_format,
            }
            if plot_format != 'html':
                save_kwargs['scale_factor'] = plot_scale_factor
            if plot_format == 'png':
                save_kwargs['ppi'] = plot_ppi
            try:
                value.save(output_path, **save_kwargs)
            except Exception as exc:
                raise PolarsTransformError(
                    f'Unable to write {plot_format.upper()} file "{output_path}": {type(exc).__name__}: {exc}'
                ) from exc
            output_kind = 'document' if plot_format == 'html' else 'image'
            return SQLResult(status=f'Wrote {plot_format.upper()} {output_kind} to {output_path}.')
        png = BytesIO()
        try:
            value.save(
                png,
                format='png',
                scale_factor=plot_scale_factor,
                ppi=plot_ppi,
            )
        except Exception as exc:
            raise PolarsTransformError(f'Unable to render Altair plot: {type(exc).__name__}: {exc}') from exc
        return SQLResult(image=png.getvalue(), image_protocol=image_protocol)
    elif value is None:
        return SQLResult()

    return SQLResult(status=f'Nothing could be displayed for return type: {type(value)}')



from __future__ import annotations

from collections.abc import Callable, Generator
import importlib
import sys
from types import ModuleType

import pytest

import mycli.packages


@pytest.fixture
def load_special(monkeypatch: pytest.MonkeyPatch) -> Generator[Callable[[bool], ModuleType], None, None]:
    original_module = sys.modules.get('mycli.packages.special')
    parent_had_special = hasattr(mycli.packages, 'special')
    original_parent_special = getattr(mycli.packages, 'special', None)

    def load(llm_off: bool) -> ModuleType:
        if llm_off:
            monkeypatch.setenv('MYCLI_LLM_OFF', '1')
        else:
            monkeypatch.delenv('MYCLI_LLM_OFF', raising=False)
        sys.modules.pop('mycli.packages.special', None)
        if hasattr(mycli.packages, 'special'):
            delattr(mycli.packages, 'special')
        return importlib.import_module('mycli.packages.special')

    yield load

    sys.modules.pop('mycli.packages.special', None)
    if original_module is not None:
        sys.modules['mycli.packages.special'] = original_module
    if parent_had_special:
        mycli.packages.special = original_parent_special  # type: ignore[attr-defined]
    elif hasattr(mycli.packages, 'special'):
        delattr(mycli.packages, 'special')


def test_special_init_exports_public_names(load_special: Callable[[bool], ModuleType]) -> None:
    special = load_special(False)

    for name in special.__all__:
        assert hasattr(special, name)


def test_special_init_reexports_special_command_api(load_special: Callable[[bool], ModuleType]) -> None:
    special = load_special(False)
    special_main = importlib.import_module('mycli.packages.special.main')

    assert special.execute is special_main.execute
    assert special.special_command is special_main.special_command
    assert special.CommandNotFound is special_main.CommandNotFound


def test_special_init_reexports_io_state_api(load_special: Callable[[bool], ModuleType]) -> None:
    special = load_special(False)
    iocommands = importlib.import_module('mycli.packages.special.iocommands')

    assert special.set_pager_enabled is iocommands.set_pager_enabled
    assert special.is_pager_enabled is iocommands.is_pager_enabled
    assert special.write_tee is iocommands.write_tee


def test_special_init_reexports_dbcommands(load_special: Callable[[bool], ModuleType]) -> None:
    special = load_special(False)
    dbcommands = importlib.import_module('mycli.packages.special.dbcommands')

    assert special.list_databases is dbcommands.list_databases
    assert special.list_tables is dbcommands.list_tables
    assert special.status is dbcommands.status


def test_special_init_uses_llm_implementation_when_enabled(load_special: Callable[[bool], ModuleType]) -> None:
    special = load_special(False)
    llm = importlib.import_module('mycli.packages.special.llm')

    assert special.FinishIteration is llm.FinishIteration
    assert special.is_llm_command is llm.is_llm_command
    assert special.handle_llm is llm.handle_llm
    assert special.sql_using_llm is llm.sql_using_llm


def test_special_init_uses_llm_stubs_when_disabled(load_special: Callable[[bool], ModuleType]) -> None:
    special = load_special(True)

    assert special.is_llm_command(r'\llm prompt') is False
    with pytest.raises(special.FinishIteration) as handle_exc:
        special.handle_llm(cast_args := object())
    with pytest.raises(special.FinishIteration) as sql_exc:
        special.sql_using_llm(cast_args)

    assert handle_exc.value.results is None
    assert sql_exc.value.results is None


def test_special_init_stub_finish_iteration_stores_results(load_special: Callable[[bool], ModuleType]) -> None:
    special = load_special(True)

    error = special.FinishIteration(results=['done'])

    assert error.results == ['done']



from __future__ import annotations

DSN_SUBCOMMANDS = {'help', 'list', 'show', 'save', 'delete'}


class DsnAliases:
    section_name: str = 'alias_dsn'

    usage = """
DSN aliases are a way to save frequently used connections
with a short alias.  You can manage them by editing ~/.myclirc
directly or by using this command.

Examples:

    # Show a DSN for the current connection
    mysql> /dsn show
    mysql://mycli@localhost/mysql

    # List all DSN aliases
    mysql> /dsn list
    ┌───────┬───────────────────────────────┐
    │ Alias │ DSN                           │
    ├───────┼───────────────────────────────┤
    │ rocks │ mysql://mycli@localhost/mysql │
    └───────┴───────────────────────────────┘

    # Save a new DSN alias based on the current connection.
    # The password will not be included!
    mysql> /dsn save connection_1

    # Delete a DSN alias.
    mysql> /dsn delete connection_1
"""

    # Class-level variable, for convenience to use as a singleton.
    instance: DsnAliases

    def __init__(self, config) -> None:
        self.config = config

    @classmethod
    def from_config(cls, config):
        return DsnAliases(config)

    def list(self) -> list[str]:
        return list(self.config.get(self.section_name, {}))

    def get(self, alias: str) -> str | None:
        return self.config.get(self.section_name, {}).get(alias, None)

    def save(self, alias: str, dsn: str) -> str:
        self.config.encoding = 'utf-8'
        if self.section_name not in self.config:
            self.config[self.section_name] = {}
        self.config[self.section_name][alias] = dsn
        self.config.write()
        return f'Saved: {alias}'

    def delete(self, alias: str) -> str:
        try:
            del self.config[self.section_name][alias]
        except KeyError:
            return f'Not Found: {alias}'
        self.config.write()
        return f'Deleted: {alias}'



from __future__ import annotations

import os
import sys
import traceback
from typing import TYPE_CHECKING, Any, Literal
from urllib.parse import quote as urlquote

import click
import keyring
import pymysql
from pymysql.constants.CR import CR_SERVER_LOST
from pymysql.constants.ER import ACCESS_DENIED_ERROR, HANDSHAKE_ERROR

from mycli.compat import WIN
from mycli.config import str_to_bool
from mycli.constants import (
    DEFAULT_CHARSET,
    DEFAULT_HOST,
    DEFAULT_PORT,
    EMPTY_PASSWORD_FLAG_SENTINEL,
    ER_MUST_CHANGE_PASSWORD_LOGIN,
)
from mycli.packages.filepaths import guess_socket_location
from mycli.packages.special.utils import format_connection_dsn
from mycli.password_sources import PasswordCandidates
from mycli.sqlexecute import SQLExecute
from mycli.ssh_tunnel import SshTunnel, SshTunnelError

try:
    from pwd import getpwuid
except ImportError:
    pass


class ClientConnectionMixin:
    if TYPE_CHECKING:
        mylogin_cnf: Any
        config: Any
        config_without_package_defaults: Any
        keepalive_ticks: int | None
        sandbox_mode: bool
        sqlexecute: Any
        logger: Any

        def read_mylogin_cnf(self, cnf: Any) -> dict[str, Any]: ...
        def echo(self, *args: Any, **kwargs: Any) -> None: ...

    def connect(
        self,
        database: str | None = "",
        user: str | None = "",
        password_candidates: PasswordCandidates | None = None,
        host: str | None = "",
        port: str | int | None = "",
        socket: str | None = "",
        character_set: str | None = "",
        local_infile: bool | None = False,
        ssl: dict[str, Any] | None = None,
        init_command: str | None = "",
        unbuffered: bool | None = None,
        use_keyring: bool | None = None,
        reset_keyring: bool | None = None,
        keepalive_ticks: int | None = None,
        ssh_jump: str | None = None,
        ssh_cli_options: str | None = None,
        vault_address: str | None = None,
        vault_mount: str | None = None,
        vault_secret: str | None = None,
        vault_password_field: str | None = None,
        vault_username_field: str | None = None,
        vault_username_from_vault: bool = False,
    ) -> None:
        mylogin_cnf: dict[str, Any] = self.read_mylogin_cnf(self.mylogin_cnf)
        # Fall back to .mylogin.cnf values only if user did not specify a value.
        user = user or mylogin_cnf["user"] or os.getenv("USER")
        host = host or mylogin_cnf["host"]
        port = port or mylogin_cnf["port"]
        ssl_config: dict[str, Any] = ssl or {}
        user_connection_config = self.config_without_package_defaults.get('connection', {})
        self.keepalive_ticks = keepalive_ticks
        self.ssh_tunnel = None
        self.selected_password = None

        int_port = port and int(port)
        if not int_port:
            int_port = DEFAULT_PORT
            if not host or host == DEFAULT_HOST:
                socket = socket or user_connection_config.get("default_socket") or mylogin_cnf["socket"] or guess_socket_location()

        if self.config.get('main', {}).get('password_sources'):
            password_source_precedence = self.config.get('main').as_list('password_sources')
        else:
            password_source_precedence = ['prompt']

        if self.config.get('main', {}).get('keyring_sources'):
            keyring_sources = self.config.get('main').as_list('keyring_sources')
        else:
            keyring_sources = []

        if ssh_jump:
            remote_host = host or DEFAULT_HOST
            remote_port = int_port or DEFAULT_PORT
            remote_socket = socket or None
            ssh_executable = self.config.get('ssh', {}).get('ssh_executable', 'ssh') or 'ssh'
            ssh_config_options = self.config.get('ssh', {}).get('ssh_options')
            tunnel_method: Literal['auto', 'socket', 'port'] = self.config.get('ssh', {}).get('tunnel_method', 'auto').lower() or 'auto'
            try:
                self.ssh_tunnel = SshTunnel.from_target(
                    ssh_jump,
                    remote_host=remote_host,
                    remote_port=int(remote_port),
                    remote_socket=remote_socket,
                    ssh_executable=ssh_executable,
                    ssh_config_options=ssh_config_options,
                    ssh_cli_options=ssh_cli_options,
                    tunnel_method=tunnel_method,
                )
                self.ssh_tunnel.start()
            except (OSError, ValueError, SshTunnelError) as exc:
                click.secho(f'Error: Unable to start SSH tunnel: {exc}', err=True, fg='red')
                try:
                    if self.ssh_tunnel:
                        self.ssh_tunnel.close()
                except Exception:
                    pass
                sys.exit(1)

        if password_candidates is None:
            password_candidates = PasswordCandidates()

        if not character_set:
            if 'main' in self.config_without_package_defaults and 'default_character_set' in self.config_without_package_defaults['main']:
                # migration with notice added with mycli 2.0.0 in 2026-07
                # todo: entirely remove support for default_character_set in [main]
                click.secho(
                    'Mycli 2.0 migration: automatically moving default_character_set from [main] to [connection] in ~/.myclirc .',
                    err=True,
                    fg='red',
                )
                character_set = self.config_without_package_defaults['main']['default_character_set']

                self.config_without_package_defaults.encoding = 'utf-8'
                if 'connection' not in self.config_without_package_defaults:
                    self.config_without_package_defaults['connection'] = {}
                if self.config_without_package_defaults['connection'].get('default_character_set', None) in (None, ''):
                    self.config_without_package_defaults['connection']['default_character_set'] = character_set
                else:
                    character_set = self.config_without_package_defaults["connection"].get("default_character_set")
                    click.secho(
                        f'But connection.default_character_set already existed, with the value: "{character_set}".',
                        err=True,
                        fg='red',
                    )
                del self.config_without_package_defaults['main']['default_character_set']
                self.config_without_package_defaults.write()

            if not character_set and 'default_character_set' in self.config['connection']:
                character_set = self.config['connection']['default_character_set']
        if not character_set:
            character_set = DEFAULT_CHARSET

        # Favor whichever local_infile option is set.
        use_local_infile = False
        for local_infile_option in (
            local_infile,
            user_connection_config.get('default_local_infile'),
            False,
        ):
            try:
                use_local_infile = str_to_bool(local_infile_option or '')
                break
            except (TypeError, ValueError):
                pass

        if 'default_ssl_ca_path' in self.config['connection'] and (not ssl_config or not ssl_config.get('capath')):
            ssl_config['capath'] = self.config['connection']['default_ssl_ca_path'] or False

        # prune lone check_hostname=False
        if not any(v for v in ssl_config.values()):
            ssl_config = {}

        ssh_tunnel_field = urlquote(ssh_jump or '')
        if ssh_tunnel_field:
            ssh_tunnel_field = ':' + ssh_tunnel_field
        keyring_identifier = f'{user}@{host}:{"" if socket else int_port}:{socket or ""}{ssh_tunnel_field}'
        keyring_domain = 'mycli.net'
        keyring_retrieved_cleanly = False

        password_candidates.add_value('login_path', mylogin_cnf['password'])
        if use_keyring and not reset_keyring:
            password_candidates.add_loader('keyring', lambda: keyring.get_password(keyring_domain, keyring_identifier))

        self.selected_password = password_candidates.resolve(password_source_precedence)
        passwd = self.selected_password.value if self.selected_password is not None else None
        keyring_retrieved_cleanly = self.selected_password is not None and self.selected_password.source == 'keyring'

        # prompt for password if requested by user
        if passwd == EMPTY_PASSWORD_FLAG_SENTINEL:
            passwd = click.prompt(f"Enter password for {user}", hide_input=True, show_default=False, default='', type=str, err=True)

        # should not fail, but will help the typechecker
        assert not isinstance(passwd, int)

        display_dsn = None
        display_dsn_user = None if vault_username_from_vault else user
        if self.ssh_tunnel:
            display_dsn = format_connection_dsn(
                user=display_dsn_user,
                host=remote_host,
                port=remote_port,
                socket=remote_socket,
                database=database,
                character_set=character_set,
                ssh_jump=ssh_jump,
                vault_address=vault_address,
                vault_mount=vault_mount,
                vault_secret=vault_secret,
                vault_password_field=vault_password_field,
                vault_username_field=vault_username_field,
            )
        elif vault_secret:
            display_dsn = format_connection_dsn(
                user=display_dsn_user,
                host=host,
                port=int_port,
                socket=socket,
                database=database,
                character_set=character_set,
                ssh_jump=ssh_jump,
                vault_address=vault_address,
                vault_mount=vault_mount,
                vault_secret=vault_secret,
                vault_password_field=vault_password_field,
                vault_username_field=vault_username_field,
            )

        connection_info: dict[str, Any] = {
            'database': database,
            'user': user,
            'password': passwd,
            'character_set': character_set,
            'local_infile': use_local_infile,
            'ssl': ssl_config,
            'init_command': init_command,
            'unbuffered': unbuffered,
            'display_dsn': display_dsn,
        }
        if self.ssh_tunnel and self.ssh_tunnel.local_socket:
            connection_info['host'] = None
            connection_info['port'] = None
            connection_info['socket'] = self.ssh_tunnel.local_socket
        elif self.ssh_tunnel:
            connection_info['host'] = self.ssh_tunnel.local_host
            connection_info['port'] = self.ssh_tunnel.local_port
            connection_info['socket'] = None
        else:
            connection_info['host'] = host
            connection_info['port'] = int_port
            connection_info['socket'] = socket

        def _update_keyring(
            password: str | None,
            password_source: str | None,
            keyring_retrieved_cleanly: bool,
        ):
            if not password:
                return
            if password_source not in keyring_sources:
                return
            if reset_keyring or (use_keyring and not keyring_retrieved_cleanly):
                try:
                    saved_pw = keyring.get_password(keyring_domain, keyring_identifier)
                    if password != saved_pw or reset_keyring:
                        keyring.set_password(keyring_domain, keyring_identifier, password)
                        click.secho(
                            f'Password from source: "{password_source}" '
                            f'saved to the system keyring at {keyring_domain}/{keyring_identifier}',
                            err=True,
                        )
                except Exception as e:
                    click.secho(f'Password not saved to the system keyring: {e}', err=True, fg='red')

        def _connect(
            retry_ssl: bool = False,
            retry_password: bool = False,
            keyring_save_eligible: bool = True,
            keyring_retrieved_cleanly: bool = False,
        ) -> None:
            try:
                if keyring_save_eligible:
                    _update_keyring(
                        connection_info["password"],
                        self.selected_password.source if self.selected_password else None,
                        keyring_retrieved_cleanly=keyring_retrieved_cleanly,
                    )
                self.sqlexecute = SQLExecute(**connection_info)
            except pymysql.OperationalError as e1:
                if e1.args[0] == HANDSHAKE_ERROR and ssl is not None and ssl.get("mode", None) == "auto":
                    # if we already tried and failed to connect without SSL, raise the error
                    if retry_ssl:
                        raise e1
                    # disable SSL and try to connect again
                    connection_info["ssl"] = None
                    _connect(
                        retry_ssl=True, keyring_retrieved_cleanly=keyring_retrieved_cleanly, keyring_save_eligible=keyring_save_eligible
                    )
                elif e1.args[0] == ACCESS_DENIED_ERROR and connection_info["password"] is None and 'fallback' in password_source_precedence:
                    # if we already tried and failed to connect with a new password, raise the error
                    if retry_password:
                        raise e1
                    password_candidates.add_loader(
                        'fallback',
                        lambda: click.prompt(
                            f"Enter password for {user}", hide_input=True, show_default=False, default='', type=str, err=True
                        ),
                    )
                    self.selected_password = password_candidates.resolve(['fallback'])
                    connection_info["password"] = self.selected_password.value if self.selected_password else None
                    keyring_retrieved_cleanly = False
                    _connect(
                        retry_password=True,
                        keyring_retrieved_cleanly=keyring_retrieved_cleanly,
                        keyring_save_eligible=keyring_save_eligible,
                    )
                elif e1.args[0] == ER_MUST_CHANGE_PASSWORD_LOGIN:
                    self.echo(
                        "Your password has expired and the server rejected the connection.",
                        err=True,
                        fg='red',
                    )
                    raise e1
                elif e1.args[0] == CR_SERVER_LOST:
                    self.echo(
                        (
                            "Connection to server lost. If this error persists, it may be a mismatch between the server and "
                            "client SSL configuration. To troubleshoot the issue, try --ssl-mode=off or --ssl-mode=on."
                        ),
                        err=True,
                        fg='red',
                    )
                    raise e1
                else:
                    raise e1

        try:
            if not WIN and socket:
                try:
                    socket_owner = getpwuid(os.stat(socket).st_uid).pw_name
                except KeyError:
                    socket_owner = '<unknown>'
                except FileNotFoundError:
                    socket_owner = '<unknown>'
                self.echo(f"Connecting to socket {socket}, owned by user {socket_owner}", err=True)
                try:
                    _connect(keyring_retrieved_cleanly=keyring_retrieved_cleanly)
                except pymysql.OperationalError as e:
                    # These are "Can't open socket" and 2x "Can't connect"
                    if [code for code in (2001, 2002, 2003) if code == e.args[0]]:
                        self.logger.debug("Database connection failed: %r.", e)
                        self.logger.error("traceback: %r", traceback.format_exc())
                        self.logger.debug("Retrying over TCP/IP")
                        self.echo(f"Failed to connect to local MySQL server through socket '{socket}':")
                        self.echo(str(e), err=True)
                        self.echo("Retrying over TCP/IP", err=True)

                        # Else fall back to TCP/IP localhost
                        socket = ""
                        host = DEFAULT_HOST
                        port = DEFAULT_PORT
                        # todo should reload the keyring identifier here instead of invalidating
                        _connect(keyring_save_eligible=False)
                    else:
                        raise e
            else:
                host = host or DEFAULT_HOST
                port = port or DEFAULT_PORT
                # could try loading the keyring again here instead of assuming nothing important changed

                # Bad ports give particularly daft error messages
                try:
                    port = int(port)
                except ValueError:
                    self.echo(f"Error: Invalid port number: '{port}'.", err=True, fg="red")
                    sys.exit(1)

                _connect(keyring_retrieved_cleanly=keyring_retrieved_cleanly)

            # Check if SQLExecute detected sandbox mode during connection
            if self.sqlexecute and self.sqlexecute.sandbox_mode:
                self.sandbox_mode = True
                self.echo(
                    "Your password has expired. Use ALTER USER or SET PASSWSORD to set a new password, or quit.",
                    err=True,
                    fg='yellow',
                )
        except Exception as e:  # Connecting to a database could fail.
            self.logger.debug("Database connection failed: %r.", e)
            self.logger.error("traceback: %r", traceback.format_exc())
            self.echo(str(e), err=True, fg="red")
            sys.exit(1)

    def reconnect(self, database: str = "") -> bool:
        """
        Attempt to reconnect to the server. Return True if successful,
        False if unsuccessful.

        The "database" argument is used only to improve messages.
        """
        assert self.sqlexecute is not None
        assert self.sqlexecute.conn is not None

        # First pass with ping(reconnect=False) and minimal feedback levels.  This definitely
        # works as expected, and is a good idea especially when "connect" was used as a
        # synonym for "use".
        try:
            self.sqlexecute.conn.ping(reconnect=False)
            if not database:
                self.echo("Already connected.", fg="yellow")
            return True
        except pymysql.err.Error:
            pass

        # Second pass with ping(reconnect=True).  It is not demonstrated that this pass ever
        # gives the benefit it is looking for, _ie_ preserves session state.  We need to test
        # this with connection pooling.
        try:
            old_connection_id = self.sqlexecute.connection_id
            self.logger.debug("Attempting to reconnect.")
            self.echo("Reconnecting...", fg="yellow")
            self.sqlexecute.conn.ping(reconnect=True)
            # if a database is currently selected, set it on the conn again
            if self.sqlexecute.dbname:
                self.sqlexecute.conn.select_db(self.sqlexecute.dbname)
            self.logger.debug("Reconnected successfully.")
            self.echo("Reconnected successfully.", fg="yellow")
            self.sqlexecute.reset_connection_id()
            if old_connection_id != self.sqlexecute.connection_id:
                self.echo("Any session state was reset.", fg="red")
            return True
        except pymysql.err.Error:
            pass

        # Third pass with sqlexecute.connect() should always work, but always resets session state.
        try:
            self.logger.debug("Creating new connection")
            self.echo("Creating new connection...", fg="yellow")
            self.sqlexecute.connect()
            self.logger.debug("New connection created successfully.")
            self.echo("New connection created successfully.", fg="yellow")
            self.echo("Any session state was reset.", fg="red")
            return True
        except pymysql.OperationalError as e:
            self.logger.debug("Reconnect failed. e: %r", e)
            self.echo(str(e), err=True, fg="red")
            return False
